# %% [markdown]
# # [Chapter 13] The final Ames sale price predictions
# 
# ## [DSLC stages]: Analysis
# 
# 
# The following code sets up the libraries and creates cleaned and pre-processed training, validation and test data that we will use in this document.
# 

# %%

#######################################
# Add new parts to run the notebook as a script
import warnings
# Filter out the specific FutureWarning about downcasting in replace
warnings.filterwarnings("ignore", category=FutureWarning, 
                        message="Downcasting behavior in `replace` is deprecated")
# from sklearn.exceptions import ConvergenceWarning
# ConvergenceWarning('ignore')
# warnings.filterwarnings('ignore')

#######################################

import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from sklearn.linear_model import LinearRegression, Ridge, Lasso
from sklego.linear_model import LADRegression
from sklearn.model_selection import cross_val_score, cross_validate
from sklearn.preprocessing import OneHotEncoder, scale, StandardScaler
from sklearn.ensemble import RandomForestRegressor
from itertools import product
from joblib import Parallel, delayed
from itertools import compress

#######################################
# Filter out scikit-learn convergence warnings
# from warnings import simplefilter
# from sklearn.exceptions import ConvergenceWarning
# simplefilter("ignore", category=ConvergenceWarning)
# warnings.filterwarnings('ignore')
# define all of the objects we need
# from prepare_ames_data import *
with open('prepare_ames_data.py', 'r') as file:
    exec(file.read())
#######################################

pd.set_option('display.max_columns', None)
pd.options.display.max_colwidth = 500
pd.options.display.max_rows = 100

# %% [markdown]
# 
# In this document we will demonstrate how to use the principles of PCS to choose the final prediction. We will demonstrate three different formats of the final prediction based on:
# 
# 1. The **single "best"** predictive algorithm, in terms of validation set performance from among a range of different algorithms each trained on several different cleaning/pre-processing judgment call-perturbed versions of the training dataset.
# 
# 1. An **ensemble** prediction, which combines the predictions from a range of predictive fits from across different algorithms and cleaning/pre-processing judgment call-perturbations that pass a predictability screening test.
# 
# 1. An **interval** of predictions from a range of predictive fits from across different algorithms and data- and cleaning/pre-processing judgment call-perturbations that pass a predictability screening test.
# 

# %% [markdown]
# 
# ## Computing the perturbed predictions
# 
# Since each of these approaches will involve each perturbed version of the cleaning/pre-processing judgment call training (and validation) datasets that we used in our stability analyses, (only the final interval approach will involve the data-perturbations), we will create the cleaning/pre-processing judgment call-perturbed datasets and fit the algorithms here.
# 
# 
# ### Create the perturbed datasets
# 
# First, let's create the list of the cleaning/pre-processing judgment call perturbed datasets:
# 

# %%
perturb_options = list(product([0.65, 0.8, 0.95], 
                               [10, 20],
                               ['other', 'mode'],
                               [True, False],
                               ['none', 'log', 'sqrt'],
                               [0, 0.5],
                               ['numeric', 'simplified_dummy', 'dummy']))
perturb_options = pd.DataFrame(perturb_options, columns=('max_identical_thresh', 
                                                         'n_neighborhoods',
                                                         'impute_missing_categorical',
                                                         'simplify_vars',
                                                         'transform_response',
                                                         'cor_feature_selection_threshold',
                                                         'convert_categorical'))

# %%
# conduct judgment call perturbations of training data
ames_jc_perturb = [preprocess_ames_data(ames_train_clean,
                                        max_identical_thresh=perturb_options['max_identical_thresh'][i],
                                        n_neighborhoods=perturb_options['n_neighborhoods'][i],
                                        impute_missing_categorical=perturb_options['impute_missing_categorical'][i],
                                        simplify_vars=perturb_options['simplify_vars'][i],
                                        transform_response=perturb_options['transform_response'][i],
                                        cor_feature_selection_threshold=perturb_options['cor_feature_selection_threshold'][i],
                                        convert_categorical=perturb_options['convert_categorical'][i])
                   for i in range(perturb_options.shape[0])]
len(ames_jc_perturb)

# %% [markdown]
# Note that several of the cleaning/pre-processing judgment call combinations will lead to identical datasets, so we will remove the duplicates from the list of perturbed datasets:

# %%
# since some judgment-call combinations yield the same data frames,
# remove duplicate data frames from the ames_jc_perturb list
new_list = []
# keep a record of which perturbations are retained
ames_jc_perturb_id = []
# add each perturbed dataset to the new_list if it isn't already included 
for i in range(len(ames_jc_perturb)):
  if not any (df.equals(ames_jc_perturb[i]) for df in new_list):
      new_list.append(ames_jc_perturb[i])
      ames_jc_perturb_id.append(i)
# update ames_jc_perturb to be this new_list object
ames_jc_perturb = new_list
# check how many perturbed datasets remain - note that this is about 40 fewer than the R version
# (presumably because I have slight differences in the implementation of the R/Python pre-processing code)
len(ames_jc_perturb)

# %%
# filter the set of perturb_options to only include the perturbations that are retained
perturb_options = perturb_options.iloc[ames_jc_perturb_id, :].reset_index(drop=True)
perturb_options.shape

# %%

# conduct judgment call perturbations of validation data data (we need to make sure each validation set is compartible with the relevant training set)
ames_val_jc_perturb = []
for i in range(len(ames_jc_perturb)):
    
    # extract relevant neighborhoods from  relevant training data
    train_neighborhood_cols = list(ames_jc_perturb[i].filter(regex="neighborhood").columns)
    train_neighborhoods = [x.replace("neighborhood_", "") for x in train_neighborhood_cols]
    
    # create preprocessed validation set
    ames_val_jc_perturb.append(
        preprocess_ames_data(ames_val_clean,
                             max_identical_thresh=perturb_options['max_identical_thresh'][i],
                             n_neighborhoods=perturb_options['n_neighborhoods'][i],
                             impute_missing_categorical=perturb_options['impute_missing_categorical'][i],
                             simplify_vars=perturb_options['simplify_vars'][i],
                             transform_response=perturb_options['transform_response'][i],
                             cor_feature_selection_threshold=perturb_options['cor_feature_selection_threshold'][i],
                             convert_categorical=perturb_options['convert_categorical'][i],
                             # make sure val set matches training set
                             column_selection=list(ames_jc_perturb[i].columns),
                             neighborhood_levels=train_neighborhoods)
        )

# create a standardized version of the validation datasets
ames_val_jc_perturb_std = []
for i in range(len(ames_val_jc_perturb)):
    df = ames_val_jc_perturb[i].drop(columns=['saleprice'])
    df_std = (df - df.mean()) / df.std()
    df_std['saleprice'] = ames_val_jc_perturb[i]['saleprice']
    ames_val_jc_perturb_std.append(df_std)



# %%
# conduct judgment call perturbations of test data data (we need to make sure each test set is compartible with the relevant training set)
ames_test_jc_perturb = []
for i in range(perturb_options.shape[0]):
    
    # extract relevant neighborhoods from  relevant training data
    train_neighborhood_cols = list(ames_jc_perturb[i].filter(regex="neighborhood").columns)
    train_neighborhoods = [x.replace("neighborhood_", "") for x in train_neighborhood_cols]
    
    # create preprocessed test set
    ames_test_jc_perturb.append(
        preprocess_ames_data(ames_test_clean,
                             max_identical_thresh=perturb_options['max_identical_thresh'][i],
                             n_neighborhoods=perturb_options['n_neighborhoods'][i],
                             impute_missing_categorical=perturb_options['impute_missing_categorical'][i],
                             simplify_vars=perturb_options['simplify_vars'][i],
                             transform_response=perturb_options['transform_response'][i],
                             cor_feature_selection_threshold=perturb_options['cor_feature_selection_threshold'][i],
                             convert_categorical=perturb_options['convert_categorical'][i],
                             # make sure test set matches training set
                             column_selection=list(ames_jc_perturb[i].columns),
                             neighborhood_levels=train_neighborhoods)
        )

# create a standardized version of the test datasets
ames_test_jc_perturb_std = []
for i in range(len(ames_test_jc_perturb)):
    df = ames_test_jc_perturb[i].drop(columns=['saleprice'])
    df_std = (df - df.mean()) / df.std()
    df_std['saleprice'] = ames_test_jc_perturb[i]['saleprice']
    ames_test_jc_perturb_std.append(df_std)


# %% [markdown]
# 
# 
# ### Fitting the algorithms to each perturbed dataset
# 
# 
# Below is a function that will fit all models simultaneously:
# 

# %%
# This code takes a while to run, so we will use the joblib library to parallelize the code
def fit_models(df):
    
    # standardize predictor variables in df for ridge and lasso
    df_x = df.drop(columns='saleprice')
    df_x_std = (df_x - df_x.mean()) / df_x.std()
    df_y = df['saleprice']
    
    ls_fit = LinearRegression().fit(X=df_x, y=df_y) 
    lad_fit = LADRegression().fit(X=df_x, y=df_y)
    rf_fit = RandomForestRegressor().fit(X=df_x, y=df_y)
        
    
    alphas = np.logspace(-1, 5, 100)
    ridge_cv_scores = []
    for alpha in alphas:
        ridge = Ridge(alpha=alpha)
        ridge_cv = cross_validate(estimator=ridge,
                                X=df_x_std,
                                y=df_y,
                                cv=10,
                                scoring='neg_root_mean_squared_error')
        ridge_cv_scores.append({'alpha': alpha,
                                'log_alpha': np.log(alpha),
                                'test_mse': -np.mean(ridge_cv['test_score'])})
        
    ridge_cv_scores_df = pd.DataFrame(ridge_cv_scores)
    ridge_alpha_min = ridge_cv_scores_df.sort_values(by='test_mse').head(1).alpha.values[0]
    # identify the 1SE value
    mse_se_ridge = ridge_cv_scores_df['test_mse'].std() / np.sqrt(10)
    mse_min_ridge = ridge_cv_scores_df['test_mse'].min()
    ridge_alpha_1se = ridge_cv_scores_df[(ridge_cv_scores_df['test_mse'] <= mse_min_ridge + mse_se_ridge) & 
                                        (ridge_cv_scores_df['test_mse'] >= mse_min_ridge - mse_se_ridge)].sort_values(by='alpha', ascending=False).head(1).alpha.values[0]
    ridge_fit = Ridge(alpha=ridge_alpha_1se).fit(X=df_x_std, y=df_y)
    
    alphas = np.logspace(-2, 7, 100)
    lasso_cv_scores = []
    for alpha in alphas:
        lasso = Lasso(alpha=alpha)
        lasso_cv = cross_validate(estimator=lasso,
                                X=df_x_std,
                                y=df_y,
                                cv=10,
                                scoring='neg_root_mean_squared_error')
        lasso_cv_scores.append({'alpha': alpha,
                                'log_alpha': np.log(alpha),
                                'test_mse': -np.mean(lasso_cv['test_score'])})
    lasso_cv_scores_df = pd.DataFrame(lasso_cv_scores)
    lasso_alpha_min = lasso_cv_scores_df.sort_values(by='test_mse').head(1).alpha.values[0]
    # identify the 1SE value
    mse_se_lasso = lasso_cv_scores_df['test_mse'].std() / np.sqrt(10)
    mse_min_lasso = lasso_cv_scores_df['test_mse'].min()
    lasso_alpha_1se = lasso_cv_scores_df[(lasso_cv_scores_df['test_mse'] <= mse_min_lasso + mse_se_lasso) & 
                                        (lasso_cv_scores_df['test_mse'] >= mse_min_lasso - mse_se_lasso)].sort_values(by='alpha', ascending=False).head(1).alpha.values[0]
    lasso_fit = Lasso(alpha=lasso_alpha_1se).fit(X=df_x_std, y=df_y)
    
    return (ls_fit, lad_fit, ridge_fit, lasso_fit, rf_fit)

# %%
results_jc = Parallel(n_jobs=-1)(delayed(fit_models)(df) for df in ames_jc_perturb)
ls_jc_perturbed, lad_jc_perturbed, ridge_jc_perturbed, lasso_jc_perturbed, rf_jc_perturbed = zip(*results_jc)

# %%
# compute the predictions on the validaion set for ls_area_perturbed, ls_multi_perturbed, ls_all_perturbed, cart_perturbed, and rf_perturbed
ls_val_jc_pred_perturbed = [ls_jc_perturbed[i].predict(X=ames_val_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_val_jc_perturb))]
lad_val_jc_pred_perturbed = [lad_jc_perturbed[i].predict(X=ames_val_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_val_jc_perturb))]
ridge_val_jc_pred_perturbed = [ridge_jc_perturbed[i].predict(X=ames_val_jc_perturb_std[i].drop(columns='saleprice')) for i in range(len(ames_val_jc_perturb_std))]
lasso_val_jc_pred_perturbed = [lasso_jc_perturbed[i].predict(X=ames_val_jc_perturb_std[i].drop(columns='saleprice')) for i in range(len(ames_val_jc_perturb_std))]
rf_val_jc_pred_perturbed = [rf_jc_perturbed[i].predict(X=ames_val_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_val_jc_perturb))]

# %%
# for predictions where the response was log-transformed, undo the log transformation
ls_val_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(ls_val_jc_pred_perturbed)]
lad_val_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(lad_val_jc_pred_perturbed)]
ridge_val_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(ridge_val_jc_pred_perturbed)]
lasso_val_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(lasso_val_jc_pred_perturbed)]
rf_val_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(rf_val_jc_pred_perturbed)]

# for predictions where the response was sqrt-transformed, undo the sqrt transformation
ls_val_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(ls_val_jc_pred_perturbed)]
lad_val_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(lad_val_jc_pred_perturbed)]
ridge_val_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(ridge_val_jc_pred_perturbed)]
lasso_val_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(lasso_val_jc_pred_perturbed)]
rf_val_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(rf_val_jc_pred_perturbed)]

# %% [markdown]
# Next, we will compute the performance of each model on the validation set for each perturbed dataset:

# %%
# LS model
ls_val_jc_perturbed_rmse = [np.sqrt(np.mean((ames_val_preprocessed['saleprice'] - ls_val_jc_pred_perturbed[i])**2)) for i in range(len(ames_val_jc_perturb))]
ls_val_jc_perturbed_mae = [np.mean(np.abs(ames_val_preprocessed['saleprice'] - ls_val_jc_pred_perturbed[i])) for i in range(len(ames_val_jc_perturb))]
ls_val_jc_perturbed_corr = [np.corrcoef(ames_val_preprocessed['saleprice'], ls_val_jc_pred_perturbed[i])[0,1] for i in range(len(ames_val_jc_perturb))]

# LAD model
lad_val_jc_perturbed_rmse = [np.sqrt(np.mean((ames_val_preprocessed['saleprice'] - lad_val_jc_pred_perturbed[i])**2)) for i in range(len(ames_val_jc_perturb))]
lad_val_jc_perturbed_mae = [np.mean(np.abs(ames_val_preprocessed['saleprice'] - lad_val_jc_pred_perturbed[i])) for i in range(len(ames_val_jc_perturb))]
lad_val_jc_perturbed_corr = [np.corrcoef(ames_val_preprocessed['saleprice'], lad_val_jc_pred_perturbed[i])[0,1] for i in range(len(ames_val_jc_perturb))]

# Ridge model
ridge_val_jc_perturbed_rmse = [np.sqrt(np.mean((ames_val_preprocessed['saleprice'] - ridge_val_jc_pred_perturbed[i])**2)) for i in range(len(ames_val_jc_perturb))]
ridge_val_jc_perturbed_mae = [np.mean(np.abs(ames_val_preprocessed['saleprice'] - ridge_val_jc_pred_perturbed[i])) for i in range(len(ames_val_jc_perturb))]
ridge_val_jc_perturbed_corr = [np.corrcoef(ames_val_preprocessed['saleprice'], ridge_val_jc_pred_perturbed[i])[0,1] for i in range(len(ames_val_jc_perturb))]

# Lasso model
lasso_val_jc_perturbed_rmse = [np.sqrt(np.mean((ames_val_preprocessed['saleprice'] - lasso_val_jc_pred_perturbed[i])**2)) for i in range(len(ames_val_jc_perturb))]
lasso_val_jc_perturbed_mae = [np.mean(np.abs(ames_val_preprocessed['saleprice'] - lasso_val_jc_pred_perturbed[i])) for i in range(len(ames_val_jc_perturb))]
lasso_val_jc_perturbed_corr = [np.corrcoef(ames_val_preprocessed['saleprice'], lasso_val_jc_pred_perturbed[i])[0,1] for i in range(len(ames_val_jc_perturb))]

# Random Forest model
rf_val_jc_perturbed_rmse = [np.sqrt(np.mean((ames_val_preprocessed['saleprice'] - rf_val_jc_pred_perturbed[i])**2)) for i in range(len(ames_val_jc_perturb))]
rf_val_jc_perturbed_mae = [np.mean(np.abs(ames_val_preprocessed['saleprice'] - rf_val_jc_pred_perturbed[i])) for i in range(len(ames_val_jc_perturb))]
rf_val_jc_perturbed_corr = [np.corrcoef(ames_val_preprocessed['saleprice'], rf_val_jc_pred_perturbed[i])[0,1] for i in range(len(ames_val_jc_perturb))]


# %%
# place all of the correlation performance lists in a data frame
perturbed_jc_corr = pd.DataFrame({
    'ls': ls_val_jc_perturbed_corr,
    'lad': lad_val_jc_perturbed_corr,
    'ridge': ridge_val_jc_perturbed_corr,
    'lasso': lasso_val_jc_perturbed_corr,
    'rf': rf_val_jc_perturbed_corr,
    'max_identical_thresh': perturb_options['max_identical_thresh'],
    'n_neighborhoods': perturb_options['n_neighborhoods'],
    'impute_missing_categorical': perturb_options['impute_missing_categorical'],
    'simplify_vars': perturb_options['simplify_vars'],
    'transform_response': perturb_options['transform_response'],
    'cor_feature_selection_threshold': perturb_options['cor_feature_selection_threshold'],
    'convert_categorical': perturb_options['convert_categorical']
    }).melt(id_vars=['max_identical_thresh', 'n_neighborhoods', 'impute_missing_categorical', 'simplify_vars', 'transform_response', 'cor_feature_selection_threshold', 'convert_categorical'], 
            var_name='model', 
            value_name='corr')

# place all of the RMSE performance lists in a data frame
perturbed_jc_rmse = pd.DataFrame({
    'ls': ls_val_jc_perturbed_rmse,
    'lad': lad_val_jc_perturbed_rmse,
    'ridge': ridge_val_jc_perturbed_rmse,
    'lasso': lasso_val_jc_perturbed_rmse,
    'rf': rf_val_jc_perturbed_rmse,
    'max_identical_thresh': perturb_options['max_identical_thresh'],
    'n_neighborhoods': perturb_options['n_neighborhoods'],
    'impute_missing_categorical': perturb_options['impute_missing_categorical'],
    'simplify_vars': perturb_options['simplify_vars'],
    'transform_response': perturb_options['transform_response'],
    'cor_feature_selection_threshold': perturb_options['cor_feature_selection_threshold'],
    'convert_categorical': perturb_options['convert_categorical']
    }).melt(id_vars=['max_identical_thresh', 'n_neighborhoods', 'impute_missing_categorical', 'simplify_vars', 'transform_response', 'cor_feature_selection_threshold', 'convert_categorical'], 
            var_name='model', 
            value_name='rmse')

# place all of the MAE performance lists in a data frame
perturbed_jc_mae = pd.DataFrame({
    'ls': ls_val_jc_perturbed_mae,
    'lad': lad_val_jc_perturbed_mae,
    'ridge': ridge_val_jc_perturbed_mae,
    'lasso': lasso_val_jc_perturbed_mae,
    'rf': rf_val_jc_perturbed_mae,
    'max_identical_thresh': perturb_options['max_identical_thresh'],
    'n_neighborhoods': perturb_options['n_neighborhoods'],
    'impute_missing_categorical': perturb_options['impute_missing_categorical'],
    'simplify_vars': perturb_options['simplify_vars'],
    'transform_response': perturb_options['transform_response'],
    'cor_feature_selection_threshold': perturb_options['cor_feature_selection_threshold'],
    'convert_categorical': perturb_options['convert_categorical']
    }).melt(id_vars=['max_identical_thresh', 'n_neighborhoods', 'impute_missing_categorical', 'simplify_vars', 'transform_response', 'cor_feature_selection_threshold', 'convert_categorical'], 
            var_name='model', 
            value_name='mae')


# %% [markdown]
# 
# ## Approach 1: Choosing a single predictive fit using PCS
# 
# Having computed the performance of each of our judgment-call perturbed fits for each algorithm we considered in this book, we can then identify which fit yields the "best" performance.
# 
# The following code prints the details of the fits with the highest correlation performance:

# %%
perturbed_jc_corr.sort_values(by='corr', ascending=False).head(6)

# %% [markdown]
# 
# Then we can print the details of the fits with the lowest rMSE (best performance):

# %%
ames_top_rmse = perturbed_jc_rmse.sort_values(by='rmse')
ames_top_rmse.head(6)

# %% [markdown]
# 
# And lastly, we can print the details of the fits with the lowest MAE (best performance):

# %%
perturbed_jc_mae.sort_values(by='mae', ascending=True).head(6)

# %% [markdown]
# 
# The "best" fit in terms of the correlation measure is the LAD algorithm with the following cleaning/pre-processing judgment call options:
# 
# - `max_identical_thresh = 0.95`
# 
# - `n_neighborhoods = 20`
# 
# - `impute_missing_categorical = "mode"`
# 
# - `simplify_vars = FALSE`
# 
# - `transform_response = "log"`
# 
# - `cor_feature_selection_threshold = 0`
# 
# - `convert_categorical = "numeric"`
# 
# (Note that these results are slightly different to the R/book version which had the square-root transformation and the dummy coding of categorical variables, but this R/book version is second-best in terms of correlation here with very little difference overall.)
# 
# The "best" fit in terms of the rMSE measure has mostly the same set of judgment calls, but involves the LAD algorithm instead of the LS algorithm and the square-root transformation instead of the logarithmic transformation, and the "best" fit in terms of the MAE involves the simplified dummy encoding of the categorical variables.
# 
# Since the rMSE and MAE measures are slightly more precise than the correlation algorithm, we will use the **LAD algorithm trained on the training set with the particular cleaning/pre-processing judgment calls corresponding to the best rMSE here (the only difference between the version here and the book version is the `convert_cateogrical` variable is "dummy" in the book but we found the best version to be "numeric" here)**
# 

# %%
ames_train_preprocessed_selected = preprocess_ames_data(ames_train_clean,
                                                        max_identical_thresh=0.95,
                                                        n_neighborhoods=20,
                                                        impute_missing_categorical='mode',
                                                        simplify_vars=False,
                                                        transform_response='sqrt',
                                                        cor_feature_selection_threshold=0,
                                                        convert_categorical='numeric')

single_fit = LADRegression()
single_fit.fit(X=ames_train_preprocessed_selected.drop(columns='saleprice'), y=ames_train_preprocessed_selected['saleprice'])

# %% [markdown]
# 
# ### Test set evaluation
# 
# Let's then evaluate this final fit using the test set (since our validation set was used to choose it, it can no longer provide an independent assessment of its performance).
# 
# First we must create the relevant pre-processed test set.

# %%
# extract relevant neighborhoods from  relevant training data
train_neighborhood_cols = ames_train_preprocessed_selected.filter(regex="neighborhood").columns
train_neighborhoods = [x.replace("neighborhood_", "") for x in train_neighborhood_cols]
    
# pre-processs the test set
ames_test_preprocessed_selected = preprocess_ames_data(ames_test_clean,
                                                        max_identical_thresh=0.95,
                                                        n_neighborhoods=20,
                                                        impute_missing_categorical='mode',
                                                        simplify_vars=False,
                                                        transform_response='sqrt',
                                                        cor_feature_selection_threshold=0,
                                                        convert_categorical='numeric',
                                                        neighborhood_levels=train_neighborhoods,
                                                        column_selection=list(ames_train_preprocessed_selected.columns))

# %% [markdown]
# And then we can compute the predictions for the test set and evaluate them.

# %%
ames_test_pred = single_fit.predict(X=ames_test_preprocessed_selected.drop(columns='saleprice'))

# %%
# print out the correlation performance
corr = np.corrcoef(ames_test_preprocessed_selected['saleprice']**2, ames_test_pred**2)[0,1]
rmse = np.sqrt(np.mean((ames_test_preprocessed_selected['saleprice']**2 - ames_test_pred**2)**2))
mae = np.mean(np.abs(ames_test_preprocessed_selected['saleprice']**2 - ames_test_pred**2))

# print out the results
print("Correlation performance:", round(corr, 3))
print("RMSE:", round(rmse, 2))
print("MAE:", round(mae, 2))



# %% [markdown]
# 
# 
# The correlation of the predicted and true test set sale prices are very high. The rMSE and MAE both indicate that the typical sale price error is less than \$20,000.

# %% [markdown]
# ## Approach 2: PCS ensemble prediction 
# 
# In this approach, we take a look at all of the predictions that we computed above (across all algorithms and judgment call combinations), and we first conduct a predictability screening test to ensure that we are not using particularly poorly performing fits to create our ensemble.
# 
# Let's visualize the distribution of the correlation performance measure across all of the algorithms and cleaning/pre-processing judgment calls (grouping by algorithm) using boxplots:

# %%
px.box(perturbed_jc_rmse, x='model', y='rmse')

# %% [markdown]
# We can also look at the distributions of the judgment calls grouping by the judgment call options, such as the response transformation:

# %%
px.box(perturbed_jc_rmse, x='transform_response', y='rmse')

# %% [markdown]
# 
# Note that the log- and square root-transformations are much more accurate in general than the fits with the untransformed response ("none") (but there are still some fits with the untransformed response that perform quite well).

# %% [markdown]
# 
# A histogram below shows the overall distribution:

# %%
px.histogram(perturbed_jc_rmse, x='rmse')

# %% [markdown]
# 
# When it comes to an ensemble fit, generally if you have a range of performance measures, you will be able to generate more accurate response predictions if you filter to just the best performing fits. Let's thus conduct a fairly arbitrary **predictability screening test of that requires a validation set correlation performance of at least 0.94**, which will filter our just some of the worse performing untransformed response fits. An ensemble prediction can then be computed based on the average prediction based on the fits that remain.

# %% [markdown]
# 
# ### Test set evaluation
# 
# To evaluate the ensemble, let's compute the ensemble predictions for each of the *test set* data points using just the fits that passed the predictability screening test.

# %% [markdown]
# Let's first compute all of the predictions for each test set house

# %%
# compute the predictions on the test set for ls_area_perturbed, ls_multi_perturbed, ls_all_perturbed, cart_perturbed, and rf_perturbed
ls_test_jc_pred_perturbed = [ls_jc_perturbed[i].predict(X=ames_test_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_test_jc_perturb))]
lad_test_jc_pred_perturbed = [lad_jc_perturbed[i].predict(X=ames_test_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_test_jc_perturb))]
ridge_test_jc_pred_perturbed = [ridge_jc_perturbed[i].predict(X=ames_test_jc_perturb_std[i].drop(columns='saleprice')) for i in range(len(ames_test_jc_perturb_std))]
lasso_test_jc_pred_perturbed = [lasso_jc_perturbed[i].predict(X=ames_test_jc_perturb_std[i].drop(columns='saleprice')) for i in range(len(ames_test_jc_perturb_std))]
rf_test_jc_pred_perturbed = [rf_jc_perturbed[i].predict(X=ames_test_jc_perturb[i].drop(columns='saleprice')) for i in range(len(ames_test_jc_perturb))]

# for predictions where the response was log-transformed, undo the log transformation
ls_test_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(ls_test_jc_pred_perturbed)]
lad_test_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(lad_test_jc_pred_perturbed)]
ridge_test_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(ridge_test_jc_pred_perturbed)]
lasso_test_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(lasso_test_jc_pred_perturbed)]
rf_test_jc_pred_perturbed = [np.exp(pred) if perturb_options['transform_response'][i] == 'log' else pred for i, pred in enumerate(rf_test_jc_pred_perturbed)]

# for predictions where the response was sqrt-transformed, undo the sqrt transformation
ls_test_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(ls_test_jc_pred_perturbed)]
lad_test_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(lad_test_jc_pred_perturbed)]
ridge_test_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(ridge_test_jc_pred_perturbed)]
lasso_test_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(lasso_test_jc_pred_perturbed)]
rf_test_jc_pred_perturbed = [pred**2 if perturb_options['transform_response'][i] == 'sqrt' else pred for i, pred in enumerate(rf_test_jc_pred_perturbed)]


# %% [markdown]
# Let's next identify which fits correspond to the top 10% of fits according to the rMSE.

# %%
# identify the 10th quantile of rMSE values across all perturbed models 
rmse_screening_threshold = perturbed_jc_rmse['rmse'].quantile(0.1)
# identify which models are in the top 10% rMSE performance band (i.e. have the lowest rMSE values)
ls_rmse_top10p = perturbed_jc_rmse.query('model == "ls"')['rmse'] <= rmse_screening_threshold
lad_rmse_top10p = perturbed_jc_rmse.query('model == "lad"')['rmse'] <= rmse_screening_threshold
ridge_rmse_top10p = perturbed_jc_rmse.query('model == "ridge"')['rmse'] <= rmse_screening_threshold
lasso_rmse_top10p = perturbed_jc_rmse.query('model == "lasso"')['rmse'] <= rmse_screening_threshold
rf_rmse_top10p = perturbed_jc_rmse.query('model == "rf"')['rmse'] <= rmse_screening_threshold

# %% [markdown]
# We can then compute the ensemble predictions using the average of all of the screened predictions across all of the models and we then aggregate the predictions from each fit into a single DataFrame

# %%
# define a concat function that can handle the cases where no models are selected
def concat_nonempty(df_list):
    if df_list == []:
        return pd.DataFrame()   
    else:
        return pd.concat([x for x in df_list if not x.empty])

# LS model
ls_test_jc_pred_perturbed_screened_df = [pd.DataFrame({
    'pred': ls_test_jc_pred_perturbed[i], 
    'model': 'ls',
    'house': list(range(len(ls_test_jc_pred_perturbed[i]))),
    'perturb_index': i
}) for i, include in enumerate(ls_rmse_top10p) if include]
ls_test_jc_pred_perturbed_screened_df = concat_nonempty(ls_test_jc_pred_perturbed_screened_df)

# LAD model
lad_test_jc_pred_perturbed_screened_df = [pd.DataFrame({
    'pred': lad_test_jc_pred_perturbed[i], 
    'model': 'lad',
    'house': list(range(len(ls_test_jc_pred_perturbed[i]))),
    'perturb_index': i
}) for i, include in enumerate(lad_rmse_top10p) if include]
lad_test_jc_pred_perturbed_screened_df = concat_nonempty(lad_test_jc_pred_perturbed_screened_df)

# Ridge model
ridge_test_jc_pred_perturbed_screened_df = [pd.DataFrame({
    'pred': ridge_test_jc_pred_perturbed[i], 
    'model': 'ridge',
    'house': list(range(len(ls_test_jc_pred_perturbed[i]))),
    'perturb_index': i
}) for i, include in enumerate(ridge_rmse_top10p) if include]
ridge_test_jc_pred_perturbed_screened_df = concat_nonempty(ridge_test_jc_pred_perturbed_screened_df)

# Lasso model
lasso_test_jc_pred_perturbed_screened_df = [pd.DataFrame({
    'pred': lasso_test_jc_pred_perturbed[i], 
    'model': 'lasso',
    'house': list(range(len(ls_test_jc_pred_perturbed[i]))),
    'perturb_index': i
}) for i, include in enumerate(lasso_rmse_top10p) if include]
lasso_test_jc_pred_perturbed_screened_df = concat_nonempty(lasso_test_jc_pred_perturbed_screened_df)

# Random Forest model
rf_test_jc_pred_perturbed_screened_df = [pd.DataFrame({
    'pred': rf_test_jc_pred_perturbed[i], 
    'model': 'rf',
    'house': list(range(len(ls_test_jc_pred_perturbed[i]))),
    'perturb_index': i
}) for i, include in enumerate(rf_rmse_top10p) if include]
rf_test_jc_pred_perturbed_screened_df = concat_nonempty(rf_test_jc_pred_perturbed_screened_df)


# %%
# aggregate the predictions from all of the models into a single data frame
test_jc_pred_perturbed_screened_df = pd.concat([
    ls_test_jc_pred_perturbed_screened_df, 
    lad_test_jc_pred_perturbed_screened_df,
    ridge_test_jc_pred_perturbed_screened_df,
    lasso_test_jc_pred_perturbed_screened_df,
    rf_test_jc_pred_perturbed_screened_df
    ])
test_jc_pred_perturbed_screened_df


# %% [markdown]
# Since this contains all of the test set predictions for all fits that passed the predictability screening test, we can then compute the ensemble predictions for each house using the average of all of the screened predictions across all of the models:

# %%
pred_ensemble_test = test_jc_pred_perturbed_screened_df.groupby('house')['pred'].mean()

# %% [markdown]
# We can then compute the performance of the ensemble predictions using the regular metrics

# %%
# compute the correlation, rmse, and mae of the ensemble predictions for the test set
test_ensemble_rmse = np.sqrt(np.mean((ames_test_preprocessed['saleprice'].values - pred_ensemble_test)**2))
test_ensemble_mae = np.mean(np.abs(ames_test_preprocessed['saleprice'].values - pred_ensemble_test))
test_ensemble_corr = np.corrcoef(ames_test_preprocessed['saleprice'].values, pred_ensemble_test)[0,1]


# %%
# print out the results
print("Correlation performance:", round(test_ensemble_corr, 3))
print("RMSE:", round(test_ensemble_rmse, 2))
print("MAE:", round(test_ensemble_mae, 2))

# %% [markdown]
# According to all three measures, the ensemble performance on the test set is slightly worse than the single-best fit performance.
# 
# 

# %% [markdown]
# 
# ## Approach 3: Calibrated PCS perturbation prediction intervals
# 
# 
# The process for computing the perturbation prediction intervals (PPIs) is similar to the ensemble prediction process, but instead of averaging the fits that pass the predictability screening test, we instead compute an interval from them. Since we want to include fits from data perturbations (bootstrap samples) when we compute the intervals, we will need to re-fit our algorithms on 10 bootstrapped versions of each of our fits that passed the (top 10% rMSE) screening stage above.
# 
# Note that only the LS and LAD fits make it into the top 10% of fits in terms of rMSE, so we will only use these two algorithms in the PPIs.

# %%
# how many LS fits made it into the top 10%?
sum(ls_rmse_top10p)

# %%
# extract just the perturbed datasets that are in the top 10% of RMSE performance for the LS model
ames_jc_perturbed_screened_ls = list(compress(ames_jc_perturb, ls_rmse_top10p))
# fit a linear regression model to each of these perturbed datasets using 10 bootstrap samples
ls_jc_perturbed_screened_boot_fit = []
for i in range(10):
    # compute a bootstrapped version of each relevant dataset
    ames_jc_perturbed_screened_ls_boot = [df.sample(n = df.shape[0], replace=True) for df in ames_jc_perturbed_screened_ls]
    # fit the LS model to each bootstrapped dataset
    for df in ames_jc_perturbed_screened_ls_boot:
        ls_jc_perturbed_screened_boot_fit.append(
            LinearRegression().fit(X=df.drop(columns='saleprice'), y=df['saleprice'])
        )

# %%
# how many LAD fits made it into the top 10%?
sum(lad_rmse_top10p)

# %%

# extract just the perturbed datasets that are in the top 10% of RMSE performance for the LAD model
ames_jc_perturbed_screened_lad = list(compress(ames_jc_perturb, lad_rmse_top10p))
# fit a linear regression model to each of these perturbed datasets using 10 bootstrap samples
lad_jc_perturbed_screened_boot_fit = []
for i in range(10):
    # compute a bootstrapped version of each relevant dataset
    ames_jc_perturbed_screened_lad_boot = [df.sample(n = df.shape[0], replace=True) for df in ames_jc_perturbed_screened_lad]
    # fit the LAD model to each bootstrapped dataset
    for df in ames_jc_perturbed_screened_lad_boot:
        lad_jc_perturbed_screened_boot_fit.append(
            LADRegression().fit(X=df.drop(columns='saleprice'), y=df['saleprice'])
        )

# %%
# how many Ridge fits made it into the top 10%?
sum(ridge_rmse_top10p)

# %%
# how many Lasso fits made it into the top 10%?
sum(lasso_rmse_top10p)

# %%
# how many RF fits made it into the top 10%?
sum(rf_rmse_top10p)

# %% [markdown]
# Next, we need to generate predictions for each of the validation set houses using each of these bootstrapped perturbed fits that passed the screening test.





import pandas as pd
import numpy as np
import logging
import sys
import os
from pathlib import Path

# Set up logging
def setup_logging(log_file=None, log_level=logging.INFO):
    """Set up logging configuration"""
    log_format = '%(asctime)s - %(levelname)s - %(message)s'
    
    # Get the root logger
    logger = logging.getLogger()
    
    # Clear any existing handlers to avoid duplicates
    if logger.handlers:
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
    
    # Configure root logger
    logger.setLevel(log_level)
    
    # Add console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter(log_format))
    logger.addHandler(console_handler)
    
    # If log file is specified, add file handler
    if log_file:
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(log_file)), exist_ok=True)
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(logging.Formatter(log_format))
        file_handler.flush = lambda: file_handler.stream.flush()
        logger.addHandler(file_handler)
        logger.info(f"Logging to file: {log_file}")
        
    return logger

# hijack sys.stdout so that every print goes through the Python logge
class StreamToLogger:
    def __init__(self, logger, level):
        self.logger = logger
        self.level  = level
        self.linebuf = ''
    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.level, line.rstrip())
    def flush(self):
        pass

sys.stdout = StreamToLogger(logging.getLogger(), logging.INFO)

# Function to format metric names nicely
def format_metric_name(metric):
    if metric == "quantile_loss":
        return "Quantile Loss"
    elif metric == "expectile_loss":
        return "Expectile Loss"
    elif metric == "picp":
        return "PICP"
    elif metric == "niw":
        return "NIW"
    elif metric == "crps":
        return "CRPS"
    elif metric == "auc":
        return "AUC"
    elif metric == "nciw":
        return "NCIW"
    elif metric == "lambda":
        return "Lambda"
    elif metric == "gamma":
        return "Gamma"
    elif metric == "mpiw":
        return "MPIW"
    elif metric == "interval_score_loss":
        return "Interval Score Loss"
    else:
        return metric.upper().replace("_", " ")

def reconstruct_dataframe(df_object):
    if isinstance(df_object, dict) and df_object.get("_pandas_dataframe_", False):
        return pd.DataFrame(
            data=df_object["data"],
            columns=df_object["columns"],
            index=df_object["index"]
        )
    return df_object

def safe_flatten(arr):
    if hasattr(arr, 'flatten'):
        return arr.flatten()
    else:
        return np.array(arr).flatten()

def get_available_datasets(models_dir: Path, coverage: int) -> list:
    """Get list of available datasets based on existing pickle files."""
    datasets = []
    # Guard: directory may not exist
    if models_dir.exists():
        # Look for pickle files with the right pattern in provided directory
        for pkl_file in models_dir.glob(f"data_*_pcs_results_{coverage}.pkl"):
            dataset_name = pkl_file.stem.replace(f"_pcs_results_{coverage}", "").replace("data_", "")
            datasets.append(dataset_name)

    # Fallback: if nothing found, search recursively from repo root for matching files
    if not datasets:
        try:
            repo_root = Path(__file__).resolve().parents[2]
            matches = list(repo_root.rglob(f"data_*_pcs_results_{coverage}.pkl"))
            for pkl_file in matches:
                dataset_name = pkl_file.stem.replace(f"_pcs_results_{coverage}", "").replace("data_", "")
                datasets.append(dataset_name)
        except Exception:
            pass

    return sorted(list(set(datasets)))

def get_top_model_info(run_data):
    """
    Get information about the top-performing model from PCS run data.
    
    Args:
        run_data: Dictionary containing run data
    
    Returns:
        (model_type, model_params, quantile_models)
    """
    logger = logging.getLogger()

    if 'top_model_names' in run_data and len(run_data['top_model_names']) > 0:
        top_model_name = run_data['top_model_names'][0]
        logger.info(f"Found top-performing model in pickle: {top_model_name}")
    else:
        if 'model_performances' in run_data and run_data['model_performances']:
            sorted_models = sorted(run_data['model_performances'].items(), 
                                key=lambda x: x[1], reverse=True)
            top_model_name = sorted_models[0][0]
            logger.info(f"Determined top model from performances: {top_model_name}")
        else:
            top_model_name = "QRF"
            logger.warning(f"Warning: No top model info found in pickle, defaulting to: {top_model_name}")
    
    model_name_map = {
        "QRF": "rf",
        "QXGB": "xgb",
        "ExpectileGAM": "qgam",
        # mean predictors are always mapped to xgb
        "OLS": "rf",
        "Ridge": "rf",
        "Lasso": "rf",
        "ElasticNet": "rf",
        "RandomForest": "rf",
        "ExtraTrees": "rf",
        "AdaBoost": "rf",
        "XGBoost": "rf",
        "MLP": "rf"
    }
    
    if top_model_name in model_name_map:
        model_type = model_name_map[top_model_name]
    else:
        for pcs_name, clear_name in model_name_map.items():
            if pcs_name.lower() == top_model_name.lower():
                model_type = clear_name
                break
        else:
            model_type = "rf"
            logger.warning(f"Warning: Unknown model type {top_model_name}, defaulting to {model_type}")

    print(f"Using top PCS model '{top_model_name}' mapped to CLEAR model type '{model_type}'")

    logger.info(f"Using top PCS model '{top_model_name}' mapped to CLEAR model type '{model_type}'")
    
    if model_type == "rf":
        model_params = {
            "n_estimators": 100,
            "random_state": 777,
            "min_samples_leaf": 10
        }
    elif model_type == "xgb":
        model_params = {
            "n_estimators": 100,
            "tree_method": "hist",
            "random_state": 777,
            "min_child_weight": 10
        }
    elif model_type == "qgam":
        try:
            if 'gam_parameters' in run_data and 'lam' in run_data['gam_parameters']:
                logger.info("Using optimal lambda from PCS: %s", run_data['gam_parameters']['lam'])
                optimal_lam = run_data['gam_parameters']['lam']
                n_splines = run_data['gam_parameters'].get('n_splines', 10)
            else:
                logger.warning("No optimal lambda found in PCS, using default value: 0.6")
                optimal_lam = 0.6
                n_splines = 10
        except Exception as e:
            logger.error(f"Error getting GAM parameters: {e}")
            optimal_lam = 0.6
            n_splines = 10
            
        model_params = {
            "n_splines": n_splines,
            "lam": optimal_lam,
            "spline_order": 3,
            "basis": "ps"
        }
    
    quantile_models = [model_type]
    
    return model_type, model_params, quantile_models


# Cleaning function for the Ames housing data
import numpy as np
import pandas as pd

# Clean the Ames data
def clean_ames_data(ames_data):
  
  ames_data_clean = ames_data.copy() 
  
  # impute missing garage year with house built year
  ames_data_clean["Garage Yr Blt"] = np.where(ames_data_clean["Garage Yr Blt"].isna(), 
                                              ames_data_clean["Garage Yr Blt"],
                                              ames_data_clean["Year Built"])
  
  # replace `Year Remod Add` value of 1950 with the `Year Built` value
  ames_data_clean["Garage Yr Blt"] = np.where(ames_data_clean["Year Remod/Add"] == 1950, 
                                              ames_data_clean["Year Built"],
                                              ames_data_clean["Year Remod/Add"])
  
  # create a date variable
  ames_data_clean["date"] = pd.to_datetime(dict(year=ames_data_clean["Yr Sold"], 
                                                month=ames_data_clean["Mo Sold"], 
                                                day=1))
  
  # clean the column names
  ames_data_clean.columns = ames_data_clean.columns.str.lower() \
                                                    .str.replace(' ', '_') \
                                                    .str.replace('/', '_')
  
  ames_data_clean = ames_data_clean.set_index("pid")
  
  return(ames_data_clean)





"""
Benchmark of real data with PCS/CQR components.

"""

import os
import sys
import pickle
import numpy as np
import pandas as pd
import argparse
import traceback
import logging
from pathlib import Path
from datetime import datetime
from sklearn.model_selection import train_test_split
import matplotlib
import time
# Import utilty functions for the experiments
from utils import safe_flatten, get_top_model_info, setup_logging, format_metric_name
# Import our scripts
script_dir = os.path.dirname(os.path.abspath(__file__))
if script_dir not in sys.path:
    sys.path.append(script_dir)
sys.path.append(os.path.join(script_dir, '..'))
from clear.metrics import evaluate_intervals
# Import the CLEAR implementation
from clear.clear import CLEAR
# Import utility functions
from clear.utils import load_ensemble_pickle

def compute_cqr_intervals(y_calib, aleatoric_median, aleatoric_lower, aleatoric_upper, 
                          pcs_median, pcs_median_test, aleatoric_median_test, aleatoric_lower_test, 
                          aleatoric_upper_test, coverage):
    """
    Compute Conformalized Quantile Regression (CQR) intervals.
    
    Args:
        y_calib: Calibration targets
        aleatoric_median/lower/upper: Aleatoric predictions for calibration data
        pcs_median: PCS median for calibration data
        pcs_median_test: PCS median for test data
        aleatoric_median/lower/upper_test: Aleatoric predictions for test data
        coverage: Target coverage probability
        
    Returns:
        (cqr_lower, cqr_upper, adjustment)
    """
    logger = logging.getLogger()
    logger.info("Computing CQR intervals...")
    
    # Center aleatoric bounds around PCS median for calibration
    calib_lower_std = pcs_median - (aleatoric_median - aleatoric_lower)
    calib_upper_std = pcs_median + (aleatoric_upper - aleatoric_median)
    
    # Calculate non-conformity scores
    scores_std = np.maximum(
        calib_lower_std - y_calib,
        y_calib - calib_upper_std
    )
    
    # Apply conformal calibration with correction term
    n_calib = len(y_calib)
    alpha = 1 - coverage
    q_level_std = (1 - alpha) * (1 + (1 / n_calib))
    q_level_std = min(q_level_std, 1.0)
    adjustment_std = np.quantile(scores_std, q_level_std, method='higher')
    
    # Generate CQR intervals using PCS median as the center
    cqr_lower = (pcs_median_test - (aleatoric_median_test - aleatoric_lower_test)) - adjustment_std
    cqr_upper = (pcs_median_test + (aleatoric_upper_test - aleatoric_median_test)) + adjustment_std
    
    return cqr_lower, cqr_upper, adjustment_std

def evaluate_and_log_metrics(all_metrics, y_test, method_name, lower_bounds, upper_bounds, alpha, evaluation_median, 
                             method_params=None):
    """
    Evaluate prediction intervals and log metrics.
    
    Args:
        all_metrics: Dictionary to update with metrics
        y_test: Test targets
        method_name: Name of the method
        lower_bounds: Lower bounds of prediction intervals
        upper_bounds: Upper bounds of prediction intervals
        alpha: Significance level (1 - coverage)
        evaluation_median: Median predictions for evaluation
        method_params: Optional parameters to store (e.g., lambda, gamma)
    
    Returns:
        Dictionary of computed metrics
    """
    metrics = evaluate_intervals(y_test, lower_bounds, upper_bounds, alpha=alpha, f=evaluation_median)
    
    # Store standard metrics
    standard_metrics = ["PICP", "NIW", "MPIW", "QuantileLoss", "ExpectileLoss", "CRPS", "AUC", "NCIW", "IntervalScoreLoss"]
    for metric in standard_metrics:
        metric_key = metric.lower()
        if metric_key not in all_metrics[method_name]:
            all_metrics[method_name][metric_key] = []
        all_metrics[method_name][metric_key].append(metrics[metric])
    
    # Store additional parameters if provided
    if method_params:
        for param_name, param_value in method_params.items():
            if param_name not in all_metrics[method_name]:
                all_metrics[method_name][param_name] = []
            all_metrics[method_name][param_name].append(param_value)
    
    return metrics

def create_metrics_dict(include_lambda=False, include_gamma=False, include_uncertainty_ratio=True):
    """Create a metrics dictionary with common metrics and optional lambda/gamma."""
    metrics = {
        "picp": [], "niw": [], "mpiw": [], "quantile_loss": [], "expectile_loss": [], "crps": [], "auc": [], "nciw": [], "interval_score_loss": []
    }
    if include_lambda:
        metrics["lambda"] = []
    if include_gamma:
        metrics["gamma"] = []
    # Add uncertainty ratio metrics
    if include_uncertainty_ratio:
        metrics["total_aleatoric_calib"] = []
        metrics["total_epistemic_calib"] = []
        metrics["uncertainty_ratio_calib"] = []
    return metrics

def generate_tables(all_metrics, dataset, coverage, results_dir, tex_dir):
    """Generate CSV tables for the given dataset and metrics"""
    # Import csv module which is needed for this function
    import csv
    logger = logging.getLogger()
    
    # Create directory if it doesn't exist
    os.makedirs(results_dir, exist_ok=True)

    # Convert coverage to a string without the decimal point
    coverage_str = str(int(coverage * 100))

    # Save individual seed results to CSV
    csv_path = os.path.join(results_dir, f"benchmark_results_{dataset}_{coverage_str}.csv")
    with open(csv_path, 'w', newline='', encoding='utf-8') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(["Dataset", "Method", "Metric", "Seed", "Value"])
        for method, metrics in all_metrics.items():
            for metric, values in metrics.items():
                for seed_idx, value in enumerate(values):
                    writer.writerow([dataset, method, metric, seed_idx, value])
    logger.info(f"CSV table saved to {os.path.abspath(csv_path)}")
    
    # LaTeX table generation has been removed as requested

def process_approach(X_train_df, y_train_df, X_valid_df, y_valid, X_test_df, y_test,
                  pcs_median_val, pcs_lower_val, pcs_upper_val,
                  pcs_median_test, pcs_lower_test, pcs_upper_test,
                  run_data, args, is_residual=False, train_pcs_median=None, 
                  pcs_lower_test_calib = None, pcs_upper_test_calib =None,
                  # Conformalized mode parameters
                  is_conformalized=False, X_calib_df=None, y_calib=None,
                  pcs_median_calib=None, pcs_lower_calib=None, pcs_upper_calib=None):
    """
    Process either the residual or non-residual approach to generate prediction intervals.
    
    Args:
        X_train/valid/test_df: Feature dataframes
        y_train/valid/test: Target values
        pcs_median/lower/upper_val: PCS bounds for validation data
        pcs_median/test: PCS bounds for test data
        run_data: Dictionary containing run data
        args: Command-line arguments
        is_residual: Whether to use the residual approach
        train_pcs_median: Training set PCS median (required for residual approach)
        
    Returns:
        Dictionary containing all computed intervals and models
    """
    logger = logging.getLogger()
    random_state = 777
    result = {}
    
    # Set approach name for logging
    approach = "residual" if is_residual else "standard"
    logger.info(f"Processing {approach} approach...")
    
    # Remove this check since we now have better handling in process_dataset
    # if (args.approach == "standard" and is_residual) or (args.approach == "residual" and not is_residual):
    #     logger.info(f"Skipping {approach} approach based on --approach setting")
    #     return {
    #         'skipped': True,
    #         'reason': 'approach_setting'
    #     }
    
    # Check if we have what we need for residual approach
    if is_residual and train_pcs_median is None:
        logger.warning("Training set PCS median required for residual approach but not provided.")
        return {
            'skipped': True,
            'reason': 'missing_train_pcs_median'
        }
        
    # Store residual flag in result for use during evaluation
    result['is_residual'] = is_residual
    
    # Define lambdas to search over - same for all models
    lambdas = np.concatenate((np.linspace(0, 0.09, 10), np.logspace(-1, 2, 4001)))
    
    logger.info(f"Using full validation set ({len(y_valid)} samples) for CLEAR calibration and baseline methods")

    # Validation data (shared across CLEAR and baseline methods)
    X_val_clear = X_valid_df
    y_val_clear = y_valid
    pcs_median_val_clear = pcs_median_val
    pcs_lower_val_clear = pcs_lower_val
    pcs_upper_val_clear = pcs_upper_val

    # Baseline methods use the same validation data
    X_val_baseline = X_valid_df
    y_val_baseline = y_valid
    pcs_median_val_baseline = pcs_median_val
    pcs_lower_val_baseline = pcs_lower_val
    pcs_upper_val_baseline = pcs_upper_val
    
    # Get top model information from the PCS run data - always use this
    model_type, model_params, quantile_models = get_top_model_info(run_data)
    
    # Create CLEAR model for aleatoric uncertainty estimation
    clear_model = create_clear_model(
        coverage=args.coverage,
        lambdas=lambdas,
        n_bootstraps=args.n_bootstraps,
        random_state=random_state,
        n_jobs=args.n_jobs
    )
    
    # Step 1bi: Fit aleatoric models using CLEAR training data (without touching calibration data)
    # Uses PCS-selected models from Step 1a and fits on train data (residuals if residual approach)
    fit_aleatoric_model(
        clear_model=clear_model,
        X_train_df=X_train_df,
        y_train_df=y_train_df,
        quantile_models=quantile_models,
        model_params=model_params,
        fit_on_residuals=is_residual,
        epistemic_preds=train_pcs_median if is_residual else None
    )
    
    # === BASELINE METHODS: Compute aleatoric predictions on full validation set ===
    logger.info(f"Computing aleatoric predictions for baseline methods on validation set ({len(y_val_baseline)} samples)...")
    aleatoric_median_val_baseline, aleatoric_lower_val_baseline, aleatoric_upper_val_baseline = clear_model.predict_aleatoric(
        X_val_baseline,
        epistemic_preds=pcs_median_val_baseline if is_residual else None
    )
    
    # === CLEAR METHODS: Compute aleatoric predictions on validation data ===
    logger.info(f"Computing aleatoric predictions for CLEAR on validation set ({len(y_val_clear)} samples)...")
    aleatoric_median_val_clear, aleatoric_lower_val_clear, aleatoric_upper_val_clear = clear_model.predict_aleatoric(
        X_val_clear,
        epistemic_preds=pcs_median_val_clear if is_residual else None
    )
    
    # Compute test predictions (same for both baseline and CLEAR)
    aleatoric_median_test, aleatoric_lower_test, aleatoric_upper_test = clear_model.predict_aleatoric(
        X_test_df,
        epistemic_preds=pcs_median_test if is_residual else None
    )
    
    # Compute calibration predictions for conformalized CLEAR (if needed)
    if is_conformalized:
        aleatoric_median_calib, aleatoric_lower_calib, aleatoric_upper_calib = clear_model.predict_aleatoric(
            X_calib_df,
            epistemic_preds=pcs_median_calib if is_residual else None
        )
    else:
        aleatoric_median_calib = aleatoric_lower_calib = aleatoric_upper_calib = None
    
    # Compute CQR intervals using baseline validation data (full validation set)
    logger.info(f"Computing CQR baseline using validation set ({len(y_val_baseline)} samples)...")
    cqr_lower, cqr_upper, adjustment = compute_cqr_intervals(
        y_calib=y_val_baseline,
        aleatoric_median=aleatoric_median_val_baseline,
        aleatoric_lower=aleatoric_lower_val_baseline,
        aleatoric_upper=aleatoric_upper_val_baseline,
        pcs_median=pcs_median_val_baseline,
        pcs_median_test=pcs_median_test,
        aleatoric_median_test=aleatoric_median_test,
        aleatoric_lower_test=aleatoric_lower_test,
        aleatoric_upper_test=aleatoric_upper_test,
        coverage=args.coverage
    )
    
    # Save CQR results
    result['cqr_lower'] = cqr_lower
    result['cqr_upper'] = cqr_upper
    result['cqr_adjustment'] = adjustment
    
    # Step 1b: Calibrate CLEAR model using validation data (without touching calibration data)
    # This implements Step 1bi (fit aleatoric models) and Step 1bii (CLEAR calibration on validation)
    logger.info("Step 1b: Calibrating CLEAR model using validation data (Step 1bi + 1bii)...")
    calibrate_clear_model(
        clear_model=clear_model,
        y_valid=y_val_clear,
        median_epistemic=pcs_median_val_clear,
        aleatoric_median=aleatoric_median_val_clear,
        aleatoric_lower=aleatoric_lower_val_clear,
        aleatoric_upper=aleatoric_upper_val_clear,
        epistemic_lower=pcs_lower_val_clear,
        epistemic_upper=pcs_upper_val_clear,
        # Conformalized CLEAR parameters
        conformalized_clear=is_conformalized,
        y_calib=y_calib if is_conformalized else None,
        median_epistemic_calib=pcs_median_calib if is_conformalized else None,
        aleatoric_median_calib=aleatoric_median_calib,
        aleatoric_lower_calib=aleatoric_lower_calib,
        aleatoric_upper_calib=aleatoric_upper_calib,
        epistemic_lower_calib=pcs_lower_calib if is_conformalized else None,
        epistemic_upper_calib=pcs_upper_calib if is_conformalized else None
    )
    
    # Step 3: Generate CLEAR predictions on unseen test data
    # Uses the calibrated model (either from Step 1b or Step 2 if conformalized)
    clear_lower, clear_upper = predict_with_clear(
        clear_model=clear_model,
        X_test_df=X_test_df,
        pcs_median_test=pcs_median_test,
        pcs_lower_test=pcs_lower_test,
        pcs_upper_test=pcs_upper_test,
        aleatoric_median_test=aleatoric_median_test, # Use the computed aleatoric median for non-conformalized CLEAR
        aleatoric_lower_test=aleatoric_lower_test,
        aleatoric_upper_test=aleatoric_upper_test
    )
    
    # Save CLEAR results
    result['clear_model'] = clear_model
    result['clear_lower'] = clear_lower
    result['clear_upper'] = clear_upper
    result['clear_optimal_lambda'] = clear_model.optimal_lambda
    result['clear_gamma'] = clear_model.gamma
    # Add uncertainty metrics
    result['clear_total_aleatoric_calib'] = clear_model.total_aleatoric_calib
    result['clear_total_epistemic_calib'] = clear_model.total_epistemic_calib
    result['clear_uncertainty_ratio_calib'] = clear_model.uncertainty_ratio_calib
    
    
    # Calculate MEAN(PCS+CQR) baseline
    mean_pcs_cqr_lower = (pcs_lower_test_calib + cqr_lower) / 2.0
    mean_pcs_cqr_upper = (pcs_upper_test_calib + cqr_upper) / 2.0
    
    # Save MEAN results
    result['mean_pcs_cqr_lower'] = mean_pcs_cqr_lower
    result['mean_pcs_cqr_upper'] = mean_pcs_cqr_upper
    
    # Implement Gamma=1 (formerly UACQR_S) method
    gamma_1_model = create_clear_model(
        coverage=args.coverage,
        lambdas=lambdas,
        n_bootstraps=args.n_bootstraps,
        random_state=random_state,
        n_jobs=args.n_jobs,
        fixed_gamma=1.0  # Fix gamma=1
    )
    
    # Calibration for Gamma=1 model
    logger.info(f"Calibrating Gamma=1 model (gamma=1, {approach} approach)")
    gamma_1_model.calibrate(
        y_calib=y_val_clear.flatten(),
        median_epistemic=pcs_median_val_clear,
        aleatoric_median=aleatoric_median_val_clear,
        aleatoric_lower=aleatoric_lower_val_clear,
        aleatoric_upper=aleatoric_upper_val_clear,
        epistemic_lower=pcs_lower_val_clear,
        epistemic_upper=pcs_upper_val_clear
    )
    logger.info(f"Gamma=1 optimal lambda: {gamma_1_model.optimal_lambda:.6f}")
    
    # Generate Gamma=1 predictions
    gamma_1_lower, gamma_1_upper = predict_with_clear(
        clear_model=gamma_1_model,
        X_test_df=X_test_df,
        pcs_median_test=pcs_median_test,
        pcs_lower_test=pcs_lower_test,
        pcs_upper_test=pcs_upper_test,
        aleatoric_median_test=aleatoric_median_test,
        aleatoric_lower_test=aleatoric_lower_test,
        aleatoric_upper_test=aleatoric_upper_test
    )
    
    # Save Gamma=1 results
    result['gamma_1_model'] = gamma_1_model
    result['gamma_1_lower'] = gamma_1_lower
    result['gamma_1_upper'] = gamma_1_upper
    result['gamma_1_lambda'] = gamma_1_model.optimal_lambda
    

    

    
    # Implement fixed lambda=1 method
    lambda_one_model = create_clear_model(
        coverage=args.coverage,
        lambdas=lambdas,
        n_bootstraps=args.n_bootstraps,
        random_state=random_state,
        n_jobs=args.n_jobs,
        fixed_lambda=1.0  # Fix lambda=1
    )
    
    # Improved calibration for Lambda=1 model - direct calibration
    logger.info(f"Calibrating Lambda=1 model ({approach} approach)")
    
    # Direct calibration to match demo_consistent.py implementation
    lambda_one_model.calibrate(
        y_calib=y_val_clear.flatten(),
        median_epistemic=pcs_median_val_clear,
        aleatoric_median=aleatoric_median_val_clear,
        aleatoric_lower=aleatoric_lower_val_clear,
        aleatoric_upper=aleatoric_upper_val_clear,
        epistemic_lower=pcs_lower_val_clear,
        epistemic_upper=pcs_upper_val_clear
    )
    
    logger.info(f"Lambda=1 gamma: {lambda_one_model.gamma:.6f}")
    
    # Generate lambda=1 predictions
    lambda_one_lower, lambda_one_upper = predict_with_clear(
        clear_model=lambda_one_model,
        X_test_df=X_test_df,
        pcs_median_test=pcs_median_test,
        pcs_lower_test=pcs_lower_test,
        pcs_upper_test=pcs_upper_test,
        aleatoric_median_test=aleatoric_median_test,
        aleatoric_lower_test=aleatoric_lower_test,
        aleatoric_upper_test=aleatoric_upper_test
    )
    
    # Save lambda=1 results
    result['lambda_one_model'] = lambda_one_model
    result['lambda_one_lower'] = lambda_one_lower
    result['lambda_one_upper'] = lambda_one_upper
    result['lambda_one_gamma'] = lambda_one_model.gamma
    
    # Implement fixed lambda=1 with conformalized CQR (lambda=1-c)
    
    # Improved calibration for Lambda=1-c model - direct calibration with conformalization
    logger.info(f"Calibrating Lambda=1-c model ({approach} approach, conformalized)")
    
    
    # Save aleatoric predictions
        # Store validation predictions (use Val1 for consistency with lambda selection)
    result['aleatoric_median_val'] = aleatoric_median_val_clear
    result['aleatoric_lower_val'] = aleatoric_lower_val_clear
    result['aleatoric_upper_val'] = aleatoric_upper_val_clear
    result['aleatoric_median_test'] = aleatoric_median_test
    result['aleatoric_lower_test'] = aleatoric_lower_test
    result['aleatoric_upper_test'] = aleatoric_upper_test
    
    # Save final median to use for evaluation
    result['evaluation_median'] = pcs_median_test
    
    return result

def process_dataset(dataset, args, run_nums=None):
    """Process a single dataset with specified parameters"""
    logger = logging.getLogger()

    # Determine dataset key and results directory based on PCS quantile model option
    dataset_key = dataset if dataset.startswith("data_") else f"data_{dataset}"
    # Base PCS results directory (point predictors by default)
    results_base = args.models_dir
    # Path to ensemble pickle file
    ensemble_file = Path(results_base) / f"{dataset_key}_pcs_results_{int(args.coverage*100)}.pkl"
    if not os.path.exists(ensemble_file):
        logger.error(f"Error: Ensemble file {ensemble_file} not found.")
        return None
    
    logger.info(f"Loading ensemble results from {ensemble_file}")
    ensemble_dict = load_ensemble_pickle(ensemble_file)
    
    # Results storage
    all_metrics = {
        "clear_vanilla": create_metrics_dict(include_lambda=True, include_gamma=True), # Was "clear"
        "clear": create_metrics_dict(include_lambda=True, include_gamma=True),         # Was "clear_residual"
        "cqr": create_metrics_dict(),
        "cqr_residual": create_metrics_dict(),
        "pcs": create_metrics_dict(),
        "a_naive": create_metrics_dict(),
        "s_naive": create_metrics_dict(),
        "mean_pcs_cqr": create_metrics_dict(),
        "mean_pcs_cqr_residual": create_metrics_dict(),
        "gamma_1": create_metrics_dict(include_lambda=True),
        "gamma_1_r": create_metrics_dict(include_lambda=True),
        "lambda_one": create_metrics_dict(include_gamma=True),
        "lambda_one_r": create_metrics_dict(include_gamma=True),
    }
    
    # Determine runs to process
    if run_nums is not None:
        runs_to_process = [f"run_{r}" for r in run_nums]
    else:
        runs_to_process = ensemble_dict.keys()
    
    # Filter to existing runs only
    runs_to_process = [run for run in runs_to_process if run in ensemble_dict]
    
    if not runs_to_process:
        logger.warning(f"No valid runs found to process for dataset {dataset}.")
        return None
    
    logger.info(f"Processing {len(runs_to_process)} runs for dataset {dataset}: {', '.join(runs_to_process)}")
    
    # Process each run
    for run_key in runs_to_process:
        start_time = time.time()
        if run_key not in ensemble_dict:
            logger.warning(f"Run {run_key} not found in ensemble pickle. Available runs: {list(ensemble_dict.keys())}")
            continue
            
        logger.info(f"\nProcessing {run_key}...")
        run_data = ensemble_dict[run_key]
        
        # Extract train, validation and test data
        X_train = run_data.get("x_train", None)
        y_train = run_data.get("y_train", None)
        X_valid = run_data.get("x_val", None)
        y_valid = run_data.get("y_val", None)
        X_test = run_data.get("x_test", None)
        y_test = run_data.get("y_test", None)
        
        # Check if we have separate calibration data (conformalized mode)
        is_conformalized = bool(run_data.get("conformalized", False))
        X_calib = run_data.get("x_calib", None)
        y_calib = run_data.get("y_calib", None)

        if is_conformalized and (X_calib is None or y_calib is None):
            logger.warning("conformalized flag is True but calibration data is missing; falling back to standard mode")
            is_conformalized = False

        if is_conformalized:
            X_calib = np.array(X_calib) if isinstance(X_calib, list) else X_calib
            y_calib = np.array(y_calib) if isinstance(y_calib, list) else y_calib
            y_calib = safe_flatten(y_calib)
            logger.info(f"Conformalized mode detected in run metadata: Train={len(y_train)}, Val={len(y_valid)}, Calib={len(y_calib)}, Test={len(y_test)}")
        else:
            # In standard mode, reuse validation as calibration
            X_calib = X_valid
            y_calib = y_valid
            logger.info(f"Standard mode (per run metadata): Train={len(y_train)}, Val/Calib={len(y_valid)}, Test={len(y_test)}")

        y_test = safe_flatten(y_test)
        y_valid = safe_flatten(y_valid)
        y_train = safe_flatten(y_train)

        # Convert to DataFrames for consistent API
        X_train_df = pd.DataFrame(X_train)
        y_train_df = pd.DataFrame(y_train, columns=['y'])
        X_valid_df = pd.DataFrame(X_valid)
        X_calib_df = pd.DataFrame(X_calib)
        X_test_df = pd.DataFrame(X_test)
        alpha = 1 - args.coverage  # Target miscoverage rate

        # print(run_data.keys())
                
        # Get validation intervals if available
        # pcs_intervals_val =np.array(run_data['val_intervals'])
        pcs_intervals_val = np.array(run_data['val_intervals_raw'])
        pcs_lower_val = pcs_intervals_val [:, 0]
        # pcs_median_val = np.array(run_data['val_median_preds'])
        pcs_median_val = pcs_intervals_val[:, 1]
        pcs_upper_val = pcs_intervals_val[:, 2]
        
        # Get calibration intervals if available (for conformalized mode)
        if is_conformalized:
            pcs_intervals_calib = np.array(run_data['calib_intervals_raw'])
            pcs_lower_calib = pcs_intervals_calib[:, 0]
            pcs_median_calib = pcs_intervals_calib[:, 1]
            pcs_upper_calib = pcs_intervals_calib[:, 2]
        else:
            # In standard mode, calibration data is same as validation
            pcs_lower_calib = pcs_lower_val
            pcs_median_calib = pcs_median_val
            pcs_upper_calib = pcs_upper_val

        # Extract the raw intervals
        pcs_intervals_test = np.array(run_data['test_intervals_raw'])
        pcs_lower_test = pcs_intervals_test[:, 0]
        pcs_median_test = pcs_intervals_test[:, 1]
        pcs_upper_test = pcs_intervals_test[:, 2]

        # Extract the calibrated intervals
        pcs_intervals_test_calib = np.asarray(run_data.get("test_intervals"))
        pcs_lower_test_calib = pcs_intervals_test_calib[:, 0]
        pcs_upper_test_calib = pcs_intervals_test_calib[:, 1]

        #  Get gamma value from pickle file
        logger.info(f"Using gamma value from pickle file: {run_data['gamma']:.4f}")
        gamma_value = run_data['gamma']

        # Check for median predictions format from train_pcs_data_quantile.py
        # Create a 2D array with the predictions as a single model
        # if isinstance(test_predictions, np.ndarray) and test_predictions.ndim == 1:
        #     test_predictions = np.array([test_predictions]).T

        # # Create a 2D array with the predictions as a single model
        # if isinstance(val_intervals, np.ndarray) and val_intervals.ndim == 1:
        #     val_intervals = np.array([val_intervals]).T
        
        # Process ensemble predictions to get PCS bounds
        logger.info("Computing PCS bounds from ensemble predictions...")

        # Validation set PCS bounds
        if isinstance(pcs_intervals_val, pd.DataFrame):
            val_ensemble_preds = pcs_intervals_val.values.T
        else:
            val_ensemble_preds = np.array(pcs_intervals_val).T
                
        # Test set PCS bounds - Properly handle different formats of test_predictions
        if isinstance(pcs_intervals_test, pd.DataFrame):
            test_ensemble_preds = pcs_intervals_test.values.T  # Shape: (n_models, n_samples)
        else:
            test_ensemble_preds = np.array(pcs_intervals_test).T  # Convert to numpy array first
        
        logger.info(f"Test ensemble predictions shape: {test_ensemble_preds.shape}")
        
        # Get train ensemble predictions for residual approach if needed
        train_ensemble_preds = None
        train_pcs_median = None
        if args.approach in ["residual", "both"]:
            pcs_intervals_train = np.array(run_data.get("train_intervals_raw"))
            if pcs_intervals_train is not None:
                if isinstance(pcs_intervals_train, pd.DataFrame):
                    train_ensemble_preds = pcs_intervals_train.values.T
                else:
                    train_ensemble_preds = np.array(pcs_intervals_train).T
                
                train_pcs_median = train_ensemble_preds[1, :]
                train_pcs_lower = train_ensemble_preds[0, :]
                train_pcs_upper = train_ensemble_preds[2, :]
                print(f"Shape of train_ensemble_preds after: {train_ensemble_preds.shape}")
            else:
                logger.warning("No training predictions found for residual approach.")
               
        try:
            # When approach is "both", process both standard and residual approaches
            if args.approach == "both":
                # First process standard approach
                standard_results = process_approach(
                    X_train_df=X_train_df, 
                    y_train_df=y_train_df,
                    X_valid_df=X_valid_df,
                    y_valid=y_valid,
                    X_test_df=X_test_df,
                    y_test=y_test,
                    pcs_median_val=pcs_median_val,
                    pcs_lower_val=pcs_lower_val,
                    pcs_upper_val=pcs_upper_val,
                    pcs_median_test=pcs_median_test,
                    pcs_lower_test=pcs_lower_test,
                    pcs_upper_test=pcs_upper_test,
                    run_data=run_data,
                    args=args,
                    is_residual=False,  # Standard approach
                    train_pcs_median=train_pcs_median,
                    pcs_lower_test_calib=pcs_lower_test_calib,
                    pcs_upper_test_calib=pcs_upper_test_calib,
                    # Conformalized mode parameters
                    is_conformalized=is_conformalized,
                    X_calib_df=X_calib_df,
                    y_calib=y_calib,
                    pcs_median_calib=pcs_median_calib,
                    pcs_lower_calib=pcs_lower_calib,
                    pcs_upper_calib=pcs_upper_calib
                )
                
                # Then process residual approach (if we have train_pcs_median)
                residual_results = None
                if train_pcs_median is not None:
                    residual_results = process_approach(
                        X_train_df=X_train_df, 
                        y_train_df=y_train_df,
                        X_valid_df=X_valid_df,
                        y_valid=y_valid,
                        X_test_df=X_test_df,
                        y_test=y_test,
                        pcs_median_val=pcs_median_val,
                        pcs_lower_val=pcs_lower_val,
                        pcs_upper_val=pcs_upper_val,
                        pcs_median_test=pcs_median_test,
                        pcs_lower_test=pcs_lower_test,
                        pcs_upper_test=pcs_upper_test,
                        run_data=run_data,
                        args=args,
                        is_residual=True,  # Residual approach
                        train_pcs_median=train_pcs_median,
                        pcs_lower_test_calib=pcs_lower_test_calib,
                        pcs_upper_test_calib=pcs_upper_test_calib,
                        # Conformalized mode parameters
                        is_conformalized=is_conformalized,
                        X_calib_df=X_calib_df,
                        y_calib=y_calib,
                        pcs_median_calib=pcs_median_calib,
                        pcs_lower_calib=pcs_lower_calib,
                        pcs_upper_calib=pcs_upper_calib
                    )
                
                # Process results from both approaches
                results_list = [r for r in [standard_results, residual_results] if r is not None]
                
                # Compute naive baselines once
                calib_predictions = pcs_median_val  # Use PCS median for calibration
                test_predictions = pcs_median_test  # Use PCS median for test
                a_naive_lower, a_naive_upper, s_naive_lower, s_naive_upper = compute_naive_intervals(
                    y_calib=y_valid,
                    calib_predictions=calib_predictions,
                    test_predictions=test_predictions,
                    coverage=args.coverage
                )
                
                # Evaluate PCS and naive baselines once
                if pcs_lower_test_calib is not None and pcs_upper_test_calib is not None:
                    evaluate_and_log_metrics(
                        all_metrics=all_metrics, y_test=y_test, method_name="pcs",
                        lower_bounds=pcs_lower_test_calib, upper_bounds=pcs_upper_test_calib,
                        alpha=alpha, evaluation_median=pcs_median_test,
                        method_params={"gamma": gamma_value} if gamma_value is not None else None
                    )
                
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name="a_naive",
                    lower_bounds=a_naive_lower, upper_bounds=a_naive_upper,
                    alpha=alpha, evaluation_median=pcs_median_test
                )
                
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name="s_naive",
                    lower_bounds=s_naive_lower, upper_bounds=s_naive_upper,
                    alpha=alpha, evaluation_median=pcs_median_test
                )
                
                # Process each set of results
                alpha = 1 - args.coverage
                for results in results_list:
                    # Skip if the approach was skipped
                    if results.get('skipped', False):
                        logger.info(f"Skipping evaluation for {run_key} because process_approach was skipped (Reason: {results.get('reason')}).")
                        continue
                        
                    # Determine evaluation median
                    evaluation_median = results.get('evaluation_median', pcs_median_test)
                    
                    # Set method prefixes based on whether this is residual or standard
                    is_res = results.get('is_residual', False)
                    # method_prefix = "clear_residual" if is_res else "clear"
                    if is_res: # Residual approach results
                        method_prefix = "clear" # Formerly "clear_residual"
                    else: # Standard approach results
                        method_prefix = "clear_vanilla" # Formerly "clear"

                    cqr_prefix = "cqr_residual" if is_res else "cqr"
                    mean_prefix = "mean_pcs_cqr_residual" if is_res else "mean_pcs_cqr"
                    gamma_1_prefix = "gamma_1_r" if is_res else "gamma_1"
                    lambda_one_prefix = "lambda_one_r" if is_res else "lambda_one"
                    
                    # Evaluate CLEAR
                    if results.get('clear_lower') is not None:
                        method_params = {
                            "lambda": results.get('clear_optimal_lambda'), 
                            "gamma": results.get('clear_gamma'),
                            "total_aleatoric_calib": results.get('clear_total_aleatoric_calib'),
                            "total_epistemic_calib": results.get('clear_total_epistemic_calib'),
                            "uncertainty_ratio_calib": results.get('clear_uncertainty_ratio_calib')
                        }
                        evaluate_and_log_metrics(
                            all_metrics=all_metrics, y_test=y_test, method_name=method_prefix,
                            lower_bounds=results['clear_lower'], upper_bounds=results['clear_upper'],
                            alpha=alpha, evaluation_median=evaluation_median,
                            method_params=method_params
                        )
                        
                        
                    # Evaluate CQR
                    if results.get('cqr_lower') is not None:
                        evaluate_and_log_metrics(
                            all_metrics=all_metrics, y_test=y_test, method_name=cqr_prefix,
                            lower_bounds=results['cqr_lower'], upper_bounds=results['cqr_upper'],
                            alpha=alpha, evaluation_median=evaluation_median
                        )

                    # Evaluate Mean(PCS+CQR)
                    if results.get('mean_pcs_cqr_lower') is not None:
                        evaluate_and_log_metrics(
                            all_metrics=all_metrics, y_test=y_test, method_name=mean_prefix,
                            lower_bounds=results['mean_pcs_cqr_lower'], upper_bounds=results['mean_pcs_cqr_upper'],
                            alpha=alpha, evaluation_median=evaluation_median
                        )

                    # Evaluate Gamma=1
                    if results.get('gamma_1_lower') is not None:
                        evaluate_and_log_metrics(
                            all_metrics=all_metrics, y_test=y_test, method_name=gamma_1_prefix,
                            lower_bounds=results['gamma_1_lower'], upper_bounds=results['gamma_1_upper'],
                            alpha=alpha, evaluation_median=evaluation_median,
                            method_params={"lambda": results.get('gamma_1_lambda')}
                        )
                    

                    # Evaluate Lambda=1
                    if results.get('lambda_one_lower') is not None:
                        evaluate_and_log_metrics(
                            all_metrics=all_metrics, y_test=y_test, method_name=lambda_one_prefix,
                            lower_bounds=results['lambda_one_lower'], upper_bounds=results['lambda_one_upper'],
                            alpha=alpha, evaluation_median=evaluation_median,
                            method_params={"gamma": results.get('lambda_one_gamma')}
                        )
                    
                # Print metrics for this run
                logger.info(f"\n{'-'*40}")
                logger.info(f"Metrics for {run_key}:")
                logger.info(f"{'-'*40}")
                
                # Define a mapping of method keys to display names
                display_name_map = {
                    "clear_vanilla": "CLEAR-Vanilla",       # Was "clear": "CLEAR"
                    "clear": "CLEAR",                 # Was "clear_residual": "CLEAR-R"
                    "cqr": "CQR",
                    "cqr_residual": "CQR-R",
                    "pcs": "PCS",
                    "a_naive": "A-Naive",
                    "s_naive": "S-Naive",
                    "mean_pcs_cqr": "Mean(PCS+CQR)",
                    "mean_pcs_cqr_residual": "Mean(PCS+CQR)-R",
                    "gamma_1": "Gamma=1",
                    "gamma_1_r": "Gamma=1-R",
                    "lambda_one": "Lambda=1",
                    "lambda_one_r": "Lambda=1-R",
                }
                
                # Log metrics for this run organized by method
                current_run_index = runs_to_process.index(run_key)
                for method_name, metrics in all_metrics.items():
                    if method_name in display_name_map:
                        # Check if this method has results for the current run
                        has_results = False
                        for metric_values in metrics.values():
                            if len(metric_values) > current_run_index:
                                has_results = True
                                break
                        
                        if has_results:
                            logger.info(f"\n  {display_name_map[method_name]}:")
                            for metric_name, values in metrics.items():
                                if len(values) > current_run_index:
                                    value = values[current_run_index]
                                    logger.info(f"    {format_metric_name(metric_name)}: {value:.4f}")
                
                # Skip the rest of the processing since we've already done it
                continue
            
            # Process only standard or residual approach based on args.approach
            results = process_approach(
                X_train_df=X_train_df, 
                y_train_df=y_train_df,
                X_valid_df=X_valid_df,
                y_valid=y_valid,
                X_test_df=X_test_df,
                y_test=y_test,
                pcs_median_val=pcs_median_val,
                pcs_lower_val=pcs_lower_val,
                pcs_upper_val=pcs_upper_val,
                pcs_median_test=pcs_median_test,
                pcs_lower_test=pcs_lower_test, # Raw PCS lower for CLEAR input
                pcs_upper_test=pcs_upper_test, # Raw PCS upper for CLEAR input
                run_data=run_data,
                args=args,
                is_residual=(args.approach == "residual"),
                train_pcs_median=train_pcs_median if train_pcs_median is not None else None,
                pcs_lower_test_calib=pcs_lower_test_calib, # Pass calibrated for Mean(PCS+CQR)
                pcs_upper_test_calib=pcs_upper_test_calib,  # Pass calibrated for Mean(PCS+CQR)
                # Conformalized mode parameters
                is_conformalized=is_conformalized,
                X_calib_df=X_calib_df,
                y_calib=y_calib,
                pcs_median_calib=pcs_median_calib,
                pcs_lower_calib=pcs_lower_calib,
                pcs_upper_calib=pcs_upper_calib
            )

            # Check if the approach was skipped (e.g., due to --residual_only)
            if results.get('skipped', False):
                logger.info(f"Skipping evaluation for {run_key} because process_approach was skipped (Reason: {results.get('reason')}).")
                continue # Skip to the next run

            # Compute naive baselines using median of raw PCS predictions
            calib_predictions = pcs_median_val # Use PCS median for calibration
            test_predictions = pcs_median_test # Use PCS median for test
            a_naive_lower, a_naive_upper, s_naive_lower, s_naive_upper = compute_naive_intervals(
                y_calib=y_valid,
                calib_predictions=calib_predictions,
                test_predictions=test_predictions,
                coverage=args.coverage
            )
            
            # Determine evaluation median
            evaluation_median = results.get('evaluation_median', pcs_median_test) # Use PCS median as fallback
            alpha = 1 - args.coverage

            # --- Start Refactored Evaluation Block ---

            # Evaluate PCS (using calibrated intervals from run_data)
            if pcs_lower_test_calib is not None and pcs_upper_test_calib is not None:
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name="pcs",
                    lower_bounds=pcs_lower_test_calib, upper_bounds=pcs_upper_test_calib,
                    alpha=alpha, evaluation_median=evaluation_median,
                    method_params={"gamma": gamma_value} if gamma_value is not None else None # gamma_value from run_data
                )
            else:
                logger.warning(f"Skipping PCS evaluation for {run_key} due to missing calibrated intervals.")

            # Evaluate Naive Baselines
            evaluate_and_log_metrics(
                all_metrics=all_metrics, y_test=y_test, method_name="a_naive",
                lower_bounds=a_naive_lower, upper_bounds=a_naive_upper,
                alpha=alpha, evaluation_median=evaluation_median
            )
            evaluate_and_log_metrics(
                all_metrics=all_metrics, y_test=y_test, method_name="s_naive",
                lower_bounds=s_naive_lower, upper_bounds=s_naive_upper,
                alpha=alpha, evaluation_median=evaluation_median
            )
                        
            # Define prefixes for metric keys based on whether residuals were used
            # Note: The results dict contains the *output* of process_approach.
            # If args.use_residual=True, these results are residual-based.
            # method_prefix = "clear_residual" if args.approach == "residual" else "clear"
            if args.approach == "residual":
                method_prefix = "clear" # Formerly "clear_residual"
            elif args.approach == "standard":
                method_prefix = "clear_vanilla" # Formerly "clear"
            else: # Should not happen if args.approach is 'both', handled above
                method_prefix = "clear_vanilla" 

            cqr_prefix = "cqr_residual" if args.approach == "residual" else "cqr"
            mean_prefix = "mean_pcs_cqr_residual" if args.approach == "residual" else "mean_pcs_cqr"
            gamma_1_prefix = "gamma_1_r" if args.approach == "residual" else "gamma_1"
            lambda_one_prefix = "lambda_one_r" if args.approach == "residual" else "lambda_one"

            # Evaluate CLEAR (Standard or Residual)
            if results.get('clear_lower') is not None:
                method_params = {
                    "lambda": results.get('clear_optimal_lambda'), 
                    "gamma": results.get('clear_gamma'),
                    "total_aleatoric_calib": results.get('clear_total_aleatoric_calib'),
                    "total_epistemic_calib": results.get('clear_total_epistemic_calib'),
                    "uncertainty_ratio_calib": results.get('clear_uncertainty_ratio_calib')
                }
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name=method_prefix,
                    lower_bounds=results['clear_lower'], upper_bounds=results['clear_upper'],
                    alpha=alpha, evaluation_median=evaluation_median,
                    method_params=method_params
                )


            # Evaluate CQR (Standard or Residual)
            if results.get('cqr_lower') is not None:
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name=cqr_prefix,
                    lower_bounds=results['cqr_lower'], upper_bounds=results['cqr_upper'],
                    alpha=alpha, evaluation_median=evaluation_median
                )

            # Evaluate Mean(PCS+CQR) (Standard or Residual)
            if results.get('mean_pcs_cqr_lower') is not None:
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name=mean_prefix,
                    lower_bounds=results['mean_pcs_cqr_lower'], upper_bounds=results['mean_pcs_cqr_upper'],
                    alpha=alpha, evaluation_median=evaluation_median
                )

            # Evaluate Gamma=1
            if results.get('gamma_1_lower') is not None:
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name=gamma_1_prefix,
                    lower_bounds=results['gamma_1_lower'], upper_bounds=results['gamma_1_upper'],
                    alpha=alpha, evaluation_median=evaluation_median,
                    method_params={"lambda": results.get('gamma_1_lambda')}
                )


            # Evaluate Lambda=1
            if results.get('lambda_one_lower') is not None:
                evaluate_and_log_metrics(
                    all_metrics=all_metrics, y_test=y_test, method_name=lambda_one_prefix,
                    lower_bounds=results['lambda_one_lower'], upper_bounds=results['lambda_one_upper'],
                    alpha=alpha, evaluation_median=evaluation_median,
                    method_params={"gamma": results.get('lambda_one_gamma')} # Store gamma if available
                )
            
            
            # Print metrics for this run
            logger.info(f"\nMetrics for {run_key}:")
            # Use the display name map defined below in the summary section
            display_name_map_local = {
                "clear_vanilla": "CLEAR-Vanilla",       # Was "clear": "CLEAR"
                "clear": "CLEAR",                 # Was "clear_residual": "CLEAR-R"
                "cqr": "CQR", "cqr_residual": "CQR-R",
                "pcs": "PCS", "a_naive": "A-Naive", "s_naive": "S-Naive",
                "mean_pcs_cqr": "Mean(PCS+CQR)", "mean_pcs_cqr_residual": "Mean(PCS+CQR)-R",
                "gamma_1": "Gamma=1", "gamma_1_r": "Gamma=1-R", 
                "lambda_one": "Lambda=1", "lambda_one_r": "Lambda=1-R",
            }
            
            # Log metrics for methods that have results for the current run
            methods_in_run = [m for m, metrics in all_metrics.items() if any(v for v in metrics.values())]
            for method in methods_in_run:
                 # Check if the method has results *for this specific run*
                has_current_run_data = False
                metrics_dict_for_log = {}
                for metric_name, values in all_metrics[method].items():
                    if len(values) == runs_to_process.index(run_key) + 1: # Check if data for current run exists
                        has_current_run_data = True
                        value = values[-1] # Get the last added value (current run)
                        metrics_dict_for_log[metric_name] = value

                if has_current_run_data:
                    display_name = display_name_map_local.get(method, method.upper())
                    logger.info(f"  {display_name}:")
                    for metric_name, value in metrics_dict_for_log.items():
                        if value is not None:
                            logger.info(f"    {format_metric_name(metric_name)}: {value:.4f}")
                        else:
                            logger.info(f"    {format_metric_name(metric_name)}: N/A")

            
            # Save results for this dataset if requested
            if args.save_results:
                # Create dataset-specific filename
                result_filename = f"{dataset}_{run_key}.pkl"
                # Ensure save_results is treated as a directory path if it doesn't end with .pkl or similar
                save_dir = args.save_results
                if not os.path.isdir(save_dir):
                     # Try to infer directory if a file path was given
                     potential_dir = os.path.dirname(save_dir)
                     if os.path.isdir(potential_dir):
                         save_dir = potential_dir
                     else:
                         # Default to a 'run_results' subdirectory if inference fails
                         save_dir = os.path.join(os.getcwd(), "run_results")
                         logger.warning(f"Specified save_results path '{args.save_results}' is not a directory. Saving to '{save_dir}'.")
                result_path = os.path.join(save_dir, result_filename)

                try:
                    run_results_save = {}
                    run_results_save['dataset'] = dataset
                    run_results_save['run'] = run_key
                    run_results_save['X_test'] = X_test
                    run_results_save['y_test'] = y_test
                    run_results_save['evaluation_median'] = evaluation_median # Include the median used

                    # --- Revised saving logic for clarity with new naming ---
                    # These are the final keys we want in the output .pkl file's metric blocks
                    # and also the keys in all_metrics.
                    target_metric_keys_map = {
                        # For standard run results:
                        "clear_vanilla": "clear_vanilla",
                        # For residual run results:
                        "clear": "clear", # This is the new "clear" (formerly clear_residual)
                        
                        # Other methods (these are processed once per run, their naming in all_metrics is already set)
                        "cqr": "cqr" if not results.get('is_residual', False) else "cqr_residual",
                        "mean_pcs_cqr": "mean_pcs_cqr" if not results.get('is_residual', False) else "mean_pcs_cqr_residual",
                        "gamma_1": "gamma_1" if not results.get('is_residual', False) else "gamma_1_r",
                        "lambda_one": "lambda_one" if not results.get('is_residual', False) else "lambda_one_r",
                        
                        # Baselines (processed once per run_key)
                        "pcs": "pcs", 
                        "a_naive": "a_naive", 
                        "s_naive": "s_naive"
                    }

                    # Keys within the 'results' dict from process_approach are generic
                    # e.g., 'clear_lower', 'gamma_1_lower'
                    # The 'is_residual' flag from 'results' tells us if these are from a standard or residual context.
                    
                    current_run_index = runs_to_process.index(run_key)
                    is_current_results_from_residual_run = results.get('is_residual', False)

                    # Loop through the methods we want to save based on the current 'results' context
                    methods_to_save_in_this_iteration = {}
                    if is_current_results_from_residual_run:
                        methods_to_save_in_this_iteration["clear"] = ("clear_lower", "clear_upper", "clear")
                        methods_to_save_in_this_iteration["mean_pcs_cqr_residual"] = ("mean_pcs_cqr_lower", "mean_pcs_cqr_upper", "mean_pcs_cqr_residual")
                        methods_to_save_in_this_iteration["gamma_1_r"] = ("gamma_1_lower", "gamma_1_upper", "gamma_1_r")
                        methods_to_save_in_this_iteration["lambda_one_r"] = ("lambda_one_lower", "lambda_one_upper", "lambda_one_r")
                    else: # Standard run
                        methods_to_save_in_this_iteration["clear_vanilla"] = ("clear_lower", "clear_upper", "clear_vanilla")
                        methods_to_save_in_this_iteration["cqr"] = ("cqr_lower", "cqr_upper", "cqr")
                        methods_to_save_in_this_iteration["mean_pcs_cqr"] = ("mean_pcs_cqr_lower", "mean_pcs_cqr_upper", "mean_pcs_cqr")
                        methods_to_save_in_this_iteration["gamma_1"] = ("gamma_1_lower", "gamma_1_upper", "gamma_1")
                        methods_to_save_in_this_iteration["lambda_one"] = ("lambda_one_lower", "lambda_one_upper", "lambda_one")

                    # Add baselines (PCS, Naive) to be saved only once, typically with non-residual/first pass
                    if not is_current_results_from_residual_run:
                        methods_to_save_in_this_iteration["pcs"] = (pcs_lower_test_calib, pcs_upper_test_calib, "pcs")
                        methods_to_save_in_this_iteration["a_naive"] = (a_naive_lower, a_naive_upper, "a_naive")
                        methods_to_save_in_this_iteration["s_naive"] = (s_naive_lower, s_naive_upper, "s_naive")
                    
                    for pkl_method_key, (res_lower_key, res_upper_key, metrics_key) in methods_to_save_in_this_iteration.items():
                        lower_val, upper_val = None, None
                        if pkl_method_key in ["pcs", "a_naive", "s_naive"]: # Already have values
                            lower_val, upper_val = res_lower_key, res_upper_key
                        else: # Fetch from 'results' dict
                            lower_val = results.get(res_lower_key)
                            upper_val = results.get(res_upper_key)

                        if lower_val is not None and upper_val is not None:
                            run_results_save[f'{pkl_method_key}_lower'] = lower_val
                            run_results_save[f'{pkl_method_key}_upper'] = upper_val

                            if metrics_key in all_metrics:
                                current_metrics_for_pkl = {}
                                for k, v_list in all_metrics[metrics_key].items():
                                    if len(v_list) > current_run_index:
                                        current_metrics_for_pkl[k] = v_list[current_run_index]
                                if current_metrics_for_pkl:
                                    run_results_save[f'{pkl_method_key}_metrics'] = current_metrics_for_pkl
                    # --- End of revised saving logic ---
                    
                    # Create directory if it doesn't exist
                    os.makedirs(os.path.dirname(result_path), exist_ok=True)
                    
                    # Save results to pickle file using binary write mode
                    with open(result_path, 'wb') as f: # Use 'wb' for pickle
                        pickle.dump(run_results_save, f)
                    
                    logger.info(f"Results saved to {os.path.abspath(result_path)}")
                
                except Exception as e:
                    logger.error(f"Error saving results for {run_key}: {e}")
                    logger.error(traceback.format_exc())
            
            
        except Exception as e:
            logger.error(f"Error processing {run_key}: {e}")
            logger.error(traceback.format_exc())
        
    
    # Print summary statistics for this dataset
    logger.info("\n" + "="*40)
    logger.info(f"Summary Statistics for {dataset} ({len(runs_to_process)} runs):")
    logger.info("="*40)
    
    # Define display names mapping (consistent with acronyms in generate_tables)
    method_display = {
        "clear_vanilla": "CLEAR-Vanilla",       # Was "clear": "CLEAR"
        "clear": "CLEAR",                 # Was "clear_residual": "CLEAR-R"
        "cqr": "CQR",
        "cqr_residual": "CQR-R",
        "pcs": "PCS",
        "a_naive": "A-Naive",
        "s_naive": "S-Naive",
        "mean_pcs_cqr": "Mean(PCS+CQR)",
        "mean_pcs_cqr_residual": "Mean(PCS+CQR)-R",
        "gamma_1": "Gamma=1",
        "gamma_1_r": "Gamma=1-R",
        "lambda_one": "Lambda=1",
        "lambda_one_r": "Lambda=1-R",
    }

    # Determine which methods have results accumulated across runs
    methods_with_results = [m for m, metrics in all_metrics.items() if any(v for v in metrics.values())]

    # Order methods for display
    ordered_methods = [
        "clear", "clear_vanilla",
        "cqr", "cqr_residual", "pcs", "a_naive", "s_naive",
        "mean_pcs_cqr", "mean_pcs_cqr_residual",
        "gamma_1", "gamma_1_r",
        "lambda_one", "lambda_one_r"
    ]
    methods_to_show = [m for m in ordered_methods if m in methods_with_results]

    # Make sure to show all metrics in a consistent order
    metric_order = ["picp", "niw", "mpiw", "quantile_loss", "expectile_loss", "crps", "interval_score_loss", "auc", "nciw", "lambda", "gamma"]
    
    for method_name in methods_to_show:
        # Skip if no display name defined (shouldn't happen with current setup)
        if method_name not in method_display: continue

        metrics = all_metrics[method_name]
        # Check if there are actually any values logged for this method
        if not any(len(v) > 0 for v in metrics.values()):
            continue
        
        logger.info(f"\n{method_display[method_name]} Metrics:")
        # Use metric_order to ensure consistent display order of metrics
        for metric_name in metric_order:
            values = metrics.get(metric_name, [])
            # values = metrics.get(metric_name)
            if len(values) > 0:  # Only display metrics that have values
            # if values: # Check if list is not empty
                avg_value = np.mean(values)
                std_value = np.std(values)
                logger.info(f"  {format_metric_name(metric_name)}: {avg_value:.4f} ± {std_value:.4f}")
                
                # If multiple runs, also print min and max values
                if len(values) > 1:
                    min_value = np.min(values)
                    max_value = np.max(values)
                    logger.info(f"    -> range: [{min_value:.4f}, {max_value:.4f}]")
    
    # Clean up large objects to release memory
    if 'ensemble_dict' in locals():
        del ensemble_dict
    import gc
    gc.collect()

    return all_metrics

def create_clear_model(coverage, lambdas, n_bootstraps, random_state, n_jobs, fixed_gamma=None, fixed_lambda=None):
    """
    Create a CLEAR model with the given parameters.
    
    Args:
        coverage: Target coverage probability
        lambdas: List of lambda values to try
        n_bootstraps: Number of bootstrap samples
        random_state: Random seed
        n_jobs: Number of parallel jobs
        fixed_gamma: Optional fixed gamma value
        fixed_lambda: Optional fixed lambda value
        
    Returns:
        Initialized CLEAR model
    """
    return CLEAR(
        desired_coverage=coverage,
        lambdas=lambdas,
        n_bootstraps=n_bootstraps,
        random_state=random_state,
        n_jobs=n_jobs,
        fixed_gamma=fixed_gamma,
        fixed_lambda=fixed_lambda
    )

def fit_aleatoric_model(clear_model, X_train_df, y_train_df, quantile_models, model_params=None, fit_on_residuals=False, epistemic_preds=None):
    """
    Fit aleatoric uncertainty model.
    
    Args:
        clear_model: CLEAR model instance
        X_train_df: Training features
        y_train_df: Training targets
        quantile_models: List of quantile models to fit
        model_params: Optional model parameters
        fit_on_residuals: Whether to fit on residuals
        epistemic_preds: Required for residual fitting
    """
    logger = logging.getLogger()
    
    if fit_on_residuals:
        if epistemic_preds is None:
            raise ValueError("epistemic_preds must be provided when fit_on_residuals=True")
        
        logger.info(f"Fitting aleatoric model on residuals with model(s): {quantile_models}")
        if model_params:
            clear_model.fit_aleatoric(
                X_train_df, y_train_df,
                quantile_model=quantile_models,
                model_params=model_params,
                fit_on_residuals=True,
                epistemic_preds=epistemic_preds
            )
        else:
            clear_model.fit_aleatoric(
                X_train_df, y_train_df,
                quantile_model=quantile_models,
                fit_on_residuals=True,
                epistemic_preds=epistemic_preds
            )
    else:
        logger.info(f"Fitting standard aleatoric model with model(s): {quantile_models}")
        if model_params:
            clear_model.fit_aleatoric(
                X_train_df, y_train_df,
                quantile_model=quantile_models,
                model_params=model_params
            )
        else:
            clear_model.fit_aleatoric(
                X_train_df, y_train_df,
                quantile_model=quantile_models
            )

def calibrate_clear_model(clear_model, y_valid, median_epistemic, aleatoric_median, 
                          aleatoric_lower, aleatoric_upper, epistemic_lower, epistemic_upper,
                          is_conformalized=False, adjustment=None,
                          conformalized_clear=False, y_calib=None, median_epistemic_calib=None,
                          aleatoric_median_calib=None, aleatoric_lower_calib=None,
                          aleatoric_upper_calib=None, epistemic_lower_calib=None,
                          epistemic_upper_calib=None):
    """
    Calibrate a CLEAR model.
    
    Args:
        clear_model: CLEAR model instance
        y_valid: Validation targets
        median_epistemic: Epistemic median predictions
        aleatoric_median: Aleatoric median predictions
        aleatoric_lower: Aleatoric lower predictions
        aleatoric_upper: Aleatoric upper predictions
        epistemic_lower: Epistemic lower predictions
        epistemic_upper: Epistemic upper predictions
        is_conformalized: Whether to use conformalized bounds
        adjustment: Conformal adjustment (required if is_conformalized=True)
        y_valid2: Val2 targets for split validation
        median_epistemic2: Val2 epistemic median
        aleatoric_median2: Val2 aleatoric median
        aleatoric_lower2: Val2 aleatoric lower
        aleatoric_upper2: Val2 aleatoric upper
        epistemic_lower2: Val2 epistemic lower
        epistemic_upper2: Val2 epistemic upper
        conformalized_clear: Whether to use conformalized CLEAR (separate cal set)
        y_calib: Calibration targets for conformalized CLEAR
        median_epistemic_calib: Calibration epistemic median for conformalized CLEAR
        aleatoric_median_calib: Calibration aleatoric median for conformalized CLEAR
        aleatoric_lower_calib: Calibration aleatoric lower for conformalized CLEAR
        aleatoric_upper_calib: Calibration aleatoric upper for conformalized CLEAR
        epistemic_lower_calib: Calibration epistemic lower for conformalized CLEAR
        epistemic_upper_calib: Calibration epistemic upper for conformalized CLEAR
    """
    logger = logging.getLogger()
    
    if is_conformalized:
        if adjustment is None:
            raise ValueError("adjustment must be provided when is_conformalized=True")
        print("-"*40)
        logger.info("Calibrating with conformalized bounds...")
        # Calculate the PCS-median-centered aleatoric widths/bounds for calibration
        calib_aleatoric_left = aleatoric_median - aleatoric_lower
        calib_aleatoric_right = aleatoric_upper - aleatoric_median

        # Apply adjustment to these centered bounds
        # The bounds are now centered around the epistemic median
        adj_lower = median_epistemic - calib_aleatoric_left - adjustment
        adj_upper = median_epistemic + calib_aleatoric_right + adjustment

        # Calibrate using the correctly centered and adjusted bounds.
        # Use median_epistemic as the aleatoric_median for calibration centering.
        clear_model.calibrate(
            y_valid.flatten(),
            median_epistemic=median_epistemic,
            aleatoric_median=median_epistemic, # Use PCS median as center
            aleatoric_lower=adj_lower,         # Pass adjusted, PCS-centered lower bound
            aleatoric_upper=adj_upper,         # Pass adjusted, PCS-centered upper bound
            epistemic_lower=epistemic_lower,
            epistemic_upper=epistemic_upper
        )
    else:
        print("-"*40)
        logger.info("Calibrating with standard bounds...")
        clear_model.calibrate(
            y_valid.flatten(),
            median_epistemic=median_epistemic,
            aleatoric_median=aleatoric_median,
            aleatoric_lower=aleatoric_lower,
            aleatoric_upper=aleatoric_upper,
            epistemic_lower=epistemic_lower,
            epistemic_upper=epistemic_upper
        )
    
    # Step 2: Conformalized CLEAR - Use separate calibration set to re-calibrate gamma
    # while keeping lambda fixed from Step 1b
    if conformalized_clear:
        # Validate that all calibration parameters are provided
        if any(v is None for v in [y_calib, median_epistemic_calib, aleatoric_median_calib,
                                   aleatoric_lower_calib, aleatoric_upper_calib,
                                   epistemic_lower_calib, epistemic_upper_calib]):
            raise ValueError("All calibration parameters must be provided when conformalized_clear=True")
        
        logger.info("Step 2: Applying conformalized CLEAR calibration using separate unseen calibration set...")
        logger.info(f"Using lambda* = {clear_model.optimal_lambda:.4f} from Step 1b, re-calibrating gamma on unseen cal set")
        
        # Step 2: Fix lambda at lambda* from Step 1b and re-calibrate gamma on the unseen calibration set
        # This uses CLEAR's built-in "lambda_fixed" calibration mode
        lambda_star = clear_model.optimal_lambda  # Store the optimized lambda from validation
        
        # Create a temporary model with fixed lambda for conformalized calibration
        clear_model.fixed_lambda = lambda_star  # Set fixed lambda
        clear_model.fixed_gamma = None  # Ensure gamma is not fixed (we want to optimize it)
        
        # Use CLEAR's built-in calibration with lambda fixed on the unseen calibration set
        clear_model.calibrate(
            y_calib=y_calib.flatten(),
            median_epistemic=median_epistemic_calib,
            aleatoric_median=aleatoric_median_calib,
            aleatoric_lower=aleatoric_lower_calib,
            aleatoric_upper=aleatoric_upper_calib,
            epistemic_lower=epistemic_lower_calib,
            epistemic_upper=epistemic_upper_calib,
            verbose=True
        )
        
        # Store conformalized parameters for reference
        clear_model.conformalized_gamma_star = clear_model.gamma  # Store gamma* from conformalization
        clear_model.conformalized_lambda_star = lambda_star  # Store lambda* used for conformalization
        
        logger.info(f"Conformalized CLEAR: lambda*={lambda_star:.4f} (fixed from Step 1b), "
                   f"gamma*={clear_model.gamma:.4f} (re-calibrated on unseen cal set)")
        logger.info(f"Calibration set size: {len(y_calib)} samples")

def predict_with_clear(clear_model, X_test_df, pcs_median_test, pcs_lower_test, pcs_upper_test, aleatoric_median_test, aleatoric_lower_test, aleatoric_upper_test):
    """
    Generate predictions with a calibrated CLEAR model.
    
    Args:
        clear_model: Calibrated CLEAR model
        X_test_df: Test features
        pcs_median_test: PCS median for test data
        pcs_lower_test: PCS lower bounds for test data
        pcs_upper_test: PCS upper bounds for test data
        aleatoric_median_test: Aleatoric median for test data
        aleatoric_lower_test: Aleatoric lower bounds for test data
        aleatoric_upper_test: Aleatoric upper bounds for test data
        
    Returns:
        (lower_bounds, upper_bounds)
    """
    return clear_model.predict(
        X_test_df,
        external_epistemic={
            'median': pcs_median_test,
            'lower': pcs_lower_test,
            'upper': pcs_upper_test
        },
        external_aleatoric={
            'median': aleatoric_median_test,
            'lower': aleatoric_lower_test,
            'upper': aleatoric_upper_test
        }
    )

def compute_naive_intervals(y_calib, calib_predictions, test_predictions, coverage):
    """
    Compute A-Naive (constant width) and S-Naive (constant absolute width) intervals.
    
    Args:
        y_calib: Calibration targets
        calib_predictions: Point predictions for calibration data
        test_predictions: Point predictions for test data
        coverage: Target coverage probability
        
    Returns:
        (a_naive_lower, a_naive_upper, s_naive_lower, s_naive_upper)
    """
    logger = logging.getLogger()
    
    # A-Naive (asymmetric constant width)
    logger.info("Computing A-Naive intervals (constant width)...")
    
    # Compute residuals on calibration data
    residuals = y_calib - calib_predictions
    
    # Find the α/2 and 1-α/2 quantiles of the residuals
    alpha = 1 - coverage
    lower_quantile = np.quantile(residuals, alpha/2, method='higher')
    upper_quantile = np.quantile(residuals, 1-alpha/2, method='higher')
    
    # Create constant-width prediction intervals
    a_naive_lower = test_predictions + lower_quantile
    a_naive_upper = test_predictions + upper_quantile
    
    # S-Naive (symmetric constant width)
    logger.info("Computing S-Naive intervals (constant absolute width)...")
    
    # Calculate absolute residuals on calibration data
    abs_residuals = np.abs(y_calib - calib_predictions)
    
    # Find the quantile corresponding to desired coverage
    gamma_naive = np.quantile(abs_residuals, coverage, method='higher')
    
    # Create fixed-width intervals
    s_naive_lower = test_predictions - gamma_naive
    s_naive_upper = test_predictions + gamma_naive
    
    return a_naive_lower, a_naive_upper, s_naive_lower, s_naive_upper


def main():
    default_datasets = [
        'data_computer',
        'data_ailerons',
        'data_airfoil',
        'data_ca_housing',
        'data_concrete',
        'data_elevator',
        'data_energy_efficiency',
        'data_insurance',
        'data_kin8nm',
        'data_miami_housing',
        'data_naval_propulsion',
        'data_parkinsons',
        'data_powerplant',
        'data_sulfur',
        'data_superconductor',
        'data_qsar',
        'data_allstate', # also takes long computationally, but we have maintained it
        ## 'data_diamond', #disabled for computational reasons
    ]
    parser = argparse.ArgumentParser(description='Inference script for CLEAR-based prediction intervals.')
    parser.add_argument("--datasets", type=str, default=','.join(default_datasets), 
                        help='Dataset key(s) to evaluate. Can be a single dataset or comma-separated list.')
    parser.add_argument("--run", type=int, help="Specific run number to process (default: process all runs).")
    parser.add_argument("--runs", type=str, help="Comma-separated list of run numbers to process (e.g., '0,1,2,3,4').")
    parser.add_argument("--coverage", type=float, default=0.95, help="Target coverage probability (default: 0.95).")
    parser.add_argument("--min_lambda", type=float, default=0, help="Minimum lambda value for grid search.")
    parser.add_argument("--max_lambda", type=float, default=100, help="Maximum lambda value for grid search.")
    parser.add_argument("--models_dir", type=str, default="../../models/pcs_top1_qpcs_10_standard", help="Directory containing ensemble results.")
    parser.add_argument("--save_results", type=str, default=None, help="Path to save results pickle file")
    parser.add_argument("--quantile_model", type=str, default="rf", 
        help="Quantile regression model to use by CLEAR. Valid options include 'linear','xgb','rf','extratrees', 'sqr', 'simultaneousquantileregressor', 'gradientboostingregressor' or a custom class.")
    parser.add_argument("--n_jobs", type=int, default=30, help="Number of jobs to run in parallel (default: 30 for high-performance systems).")
    parser.add_argument("--n_bootstraps", type=int, default=100, help="Number of bootstrap samples (default: 100)")
    parser.add_argument("--generate_tables", action="store_true", help="Generate CSV tables for aggregated metrics.")
    # Add new argument for median source
    parser.add_argument("--median_source", type=str, choices=["pcs","cqr"], default="pcs", help="Source for point predictor baseline: 'pcs' for ensemble median or 'cqr' for quantile regression median.")
    # Add logging-related arguments
    parser.add_argument("--log_file", type=str, default="auto", 
                        help="Path to log file. Use 'none' to disable file logging. Default 'auto' creates timestamped logs.")
    parser.add_argument("--global_log", action="store_true",
                        help="Create a single log file for all datasets rather than separate files.")
    parser.add_argument("--log_level", type=str, default="INFO", 
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                        help="Logging level.")
    parser.add_argument("--approach", type=str, choices=["standard", "residual", "both"], 
                    default="standard", help="Approach to use: standard, residual, or both")
    parser.add_argument("--csv_results_dir", type=str, default=None,
                        help="Directory where the CSV table files will be saved. Can be an absolute path or "
                             "relative to the current working directory. If not provided, defaults to "
                             "'[PROJECT_ROOT]/results'. Example: '../../results/my_variant_results'.")
    args = parser.parse_args()

    # Remove the validation check for residual_only
    # if args.residual_only and not args.use_residual:
    #     print("Error: --residual_only requires --use_residual to be enabled.")
    #     return

    # Set up logging
    log_level = getattr(logging, args.log_level)
    
    # Base directory is two levels up from the script
    base_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    logs_dir = os.path.join(base_dir, "logs")
    os.makedirs(logs_dir, exist_ok=True)
    
    # Global log setup - create a single log file for all datasets
    if args.global_log and args.log_file.lower() != "none":
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        
        # Get all datasets as a single string
        datasets = "_".join([d.strip() for d in args.datasets.split(",")])
        if len(datasets) > 30:  # Truncate if too long
            datasets = datasets[:27] + "..."
            
        # Format a descriptive filename
        coverage = f"{int(args.coverage*100)}"
        
        # Use custom name or auto-generated
        if args.log_file.lower() != "auto":
            base_name = os.path.splitext(args.log_file)[0]
            global_log_file = os.path.join(logs_dir, f"{base_name}_{timestamp}.log")
        else:
            global_log_file = os.path.join(logs_dir, f"benchmark_all_{timestamp}.log")
        
        # Set up the global logger
        logger = setup_logging(global_log_file, log_level)
        logger.info(f"Starting global log for all experiments at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
        logger.info(f"Datasets: {args.datasets}")
        logger.info(f"Command-line arguments: {args}")
        
        # Store the global log file
        args.current_log_file = global_log_file
    else:
        # No global log, will set up individual loggers for each dataset
        args.current_log_file = None
        logger = logging.getLogger()
        logger.setLevel(log_level)

    # Check for mutually exclusive arguments
    if args.run is not None and args.runs is not None:
        logger.error("Error: Cannot specify both --run and --runs. Please use only one.")
        return

    # Define output directory for CSV tables
    script_dir = os.path.dirname(os.path.abspath(__file__))
    base_dir = os.path.dirname(os.path.dirname(script_dir)) # base_dir is the project root
    
    if args.csv_results_dir:
        # User-provided path, resolve it (can be relative to CWD or absolute)
        tables_csv_dir = os.path.abspath(args.csv_results_dir)
    else:
        # Default path relative to project root
        tables_csv_dir = os.path.join(base_dir, "results")
    
    # Create directory if it doesn't exist
    os.makedirs(tables_csv_dir, exist_ok=True)
    
    # Determine run numbers to process
    run_nums = None
    if args.run is not None:
        run_nums = [args.run]
    elif args.runs is not None:
        run_nums = [int(r.strip()) for r in args.runs.split(",")]
    
    # Process datasets
    datasets = [d.strip() for d in args.datasets.split(",")]
    all_dataset_metrics = {}
    
    for dataset in datasets:
        # Set up individual logging for this dataset if we're not using a global log
        if not args.global_log and args.log_file.lower() != "none":
            # Create a descriptive filename based on the command
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            coverage = f"{int(args.coverage*100)}"
            
            # Include run numbers in filename if specified
            run_part = ""
            if args.run is not None:
                run_part = f"_run{args.run}"
            elif args.runs is not None:
                run_nums = [r.strip() for r in args.runs.split(",")]
                run_part = f"_runs{'_'.join(run_nums)}"
            
            # If user specified a custom log_file name that's not "auto", use it as a base name
            if args.log_file.lower() != "auto":
                # Strip extension if present
                base_name = os.path.splitext(args.log_file)[0]
                log_file = os.path.join(logs_dir, f"{base_name}_{dataset}_{timestamp}.log")
            else:
                # Use auto-generated name with dataset, coverage, and timestamp
                log_file = os.path.join(logs_dir, f"benchmark_{dataset}_cov{coverage}{run_part}_{timestamp}.log")
                
            # Set up dataset-specific logger
            logger = setup_logging(log_file, log_level)
            logger.info(f"Logging dataset {dataset} to file: {log_file}")
        else:
            # Use the global logger or console-only logger already set up
            logger = logging.getLogger()
            if args.current_log_file:
                logger.info(f"\n{'-'*80}\nProcessing dataset: {dataset}\n{'-'*80}")
        
        logger.info(f"\n{'='*40}\nProcessing dataset: {dataset}\n{'='*40}")
        
        try:
            # Process the dataset and get metrics
            dataset_metrics = process_dataset(dataset, args, run_nums)
            
            if dataset_metrics:
                all_dataset_metrics[dataset] = dataset_metrics
                
                # Generate tables for this dataset if requested
                if args.generate_tables:
                    generate_tables(dataset_metrics, dataset, args.coverage, tables_csv_dir, None)
        except Exception as e:
            logger.error(f"Error processing dataset {dataset}: {e}")
            logger.error(traceback.format_exc())
        
        # Force garbage collection between datasets to free memory
        import gc
        gc.collect()
    
    logger.info("\nBenchmark completed!")
    
if __name__ == "__main__":
    # Set the plot rendering to non-interactive
    matplotlib.use('Agg')
    main()


###################################################################
# Standard version
# To run the experiments with the models for `qpcs_10` (variant a) then followed by `qxgb_10` (variant b) and `pcs_10` (variant c), use the following command:
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 25 --global_log --approach both --models_dir ../../models/pcs_top1_qpcs_10_standard --csv_results_dir ../../results/standard/qPCS_all_10seeds_all
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_qxgb_10_standard --csv_results_dir ../../results/standard/qPCS_qxgb_10seeds_qxgb
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_pcs_10_standard --csv_results_dir ../../results/standard/PCS_all_10seeds_qrf
# Note that the GAM in variant a may not sometimes converge, in which case just reduce the number of parallel cores. Sometimes, expectileGAM may fail to converge. See the error below:
# https://github.com/dswah/pyGAM/issues/357

###################################################################
# Conformalized version
# To run the experiments with the models for `qpcs_10` (variant a) then followed by `qxgb_10` (variant b) and `pcs_10` (variant c), use the following command:
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 25 --global_log --approach both --models_dir ../../models/pcs_top1_qpcs_10_conformalized --csv_results_dir ../../results/conformalized/qPCS_all_10seeds_all
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_qxgb_10_conformalized --csv_results_dir ../../results/conformalized/qPCS_qxgb_10seeds_qxgb
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_pcs_10_conformalized --csv_results_dir ../../results/conformalized/PCS_all_10seeds_qrf
# Note that the GAM in variant a may not sometimes converge, in which case just reduce the number of parallel cores. Sometimes, expectileGAM may fail to converge. See the error below:
# https://github.com/dswah/pyGAM/issues/357

###################################################################
# One liner to do variant b, c and a (fastest order)

## Standard version
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 25 --global_log --approach both --models_dir ../../models/pcs_top1_qxgb_10_standard --csv_results_dir ../../results/standard/qPCS_qxgb_10seeds_qxgb ; python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_pcs_10_standard --csv_results_dir ../../results/standard/PCS_all_10seeds_qrf ; python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_qpcs_10_standard --csv_results_dir ../../results/standard/qPCS_all_10seeds_all

## Conformalized version
# python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 25 --global_log --approach both --models_dir ../../models/pcs_top1_qxgb_10_conformalized --csv_results_dir ../../results/conformalized/qPCS_qxgb_10seeds_qxgb ; python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_pcs_10_conformalized --csv_results_dir ../../results/conformalized/PCS_all_10seeds_qrf ; python benchmark_real_data.py --coverage 0.95 --generate_tables --n_jobs 30 --global_log --approach both --models_dir ../../models/pcs_top1_qpcs_10_conformalized --csv_results_dir ../../results/conformalized/qPCS_all_10seeds_all


import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from copy import deepcopy
import os
import sys
import argparse
import pickle
import time
import logging
# Define models - exactly as in original
from sklearn.linear_model import RidgeCV, LassoCV, ElasticNetCV
from xgboost import XGBRegressor
from joblib import Parallel, delayed
from quantile_forest import RandomForestQuantileRegressor, ExtraTreesQuantileRegressor
from pygam import ExpectileGAM
from utils import convert_to_serializable, get_intervals_manually, optimize_gam_smoothing, setup_logging, print_model_performance, WorkerLogger

# Update imports to use the new PCS implementation
# Try to import from the installed package first, fall back to local path if not available
try:
    from PCS.regression import PCS_UQ, PCS_OOB
except ImportError:
    # Add PCS_UQ to path if needed
    pcs_uq_path = os.path.join("..", "..", "PCS_UQ", "src")
    if pcs_uq_path not in sys.path:
        sys.path.append(pcs_uq_path)
    from PCS.regression import PCS_UQ, PCS_OOB

def run_single_experiment_pcs(dataset_name, seed_index, n_boot=50, base_data_path=None, alphas=[0.1, 0.05, 0.01], top_k=1, use_oob=False, calibration_method='multiplicative', conformalized=False, qxgb_only=False):
    """
    Run a single PCS ensemble experiment for a dataset using the specified seed index.
    
    Args:
        dataset_name: Name of the dataset to process (e.g., 'data_ailerons' or just 'ailerons')
        seed_index: The index to use for generating the seed (seed = 777 + seed_index)
        n_boot: Number of bootstrap samples
        base_data_path: Base directory path where data is located
        alphas: List of alpha values to use for calibration (default: [0.1, 0.05, 0.01])
        top_k: Number of top-performing models to use (default: 1)
        use_oob: Whether to use PCS_OOB instead of PCS_UQ (default: False)
        calibration_method: Method for calibration, 'multiplicative' or 'additive' (default: 'multiplicative')
        conformalized: Whether to use separate calibration set (conformalized) or reuse validation set (standard) (default: False)
        qxgb_only: If True, use only XGBoost quantile estimator (default: False)
    
    Returns:
        Dict containing model results and trained models for each alpha value
    """
    # Create a worker logger to capture logs
    worker_logger = WorkerLogger()
    
    # Normalize base path
    if base_data_path is None:
        base_data_path = os.getcwd()
    
    # Make sure dataset name has 'data_' prefix
    if not dataset_name.startswith("data_"):
        dataset_name = f"data_{dataset_name}"
    
    # List all possible paths where the data might be
    search_paths = [
        # Your current pwd shows you're in F:\research\clear\data
        os.path.join(base_data_path, dataset_name),
        os.path.join(base_data_path, dataset_name),
        # Try parent directories
        os.path.join(os.path.dirname(base_data_path), dataset_name),
        os.path.join(os.path.dirname(base_data_path), "data", dataset_name)
    ]
    
    # Find an existing path that contains the dataset
    data_dir = next((path for path in search_paths if os.path.exists(path)), None)
    
    if not data_dir:
        worker_logger.log(f"Searched for dataset in these directories: {search_paths}")
        raise FileNotFoundError(f"Could not find directory for {dataset_name}")
    
    # Now construct paths to the X and y files
    X_path = os.path.join(data_dir, "X.csv")
    y_path = os.path.join(data_dir, "y.csv")
    
    # Verify files exist
    if not os.path.exists(X_path) or not os.path.exists(y_path):
        worker_logger.log(f"Found directory {data_dir} but missing X.csv or y.csv files")
        raise FileNotFoundError(f"Missing data files in {data_dir}")
    
    worker_logger.log(f"Using X data from: {X_path}")
    worker_logger.log(f"Using y data from: {y_path}")
    
    # Read the data
    X = pd.read_csv(X_path)
    y = np.loadtxt(y_path, delimiter=',')
    
    # Calculate seed from index - exactly as in original
    seed = 777 + seed_index
    
    # Train, validation, and test split
    x_train_val, x_test, y_train_val, y_test = train_test_split(X, y, test_size=0.2, random_state=seed)
    
    if conformalized:
        # For conformalized: split train_val into train (75%), val (12.5%), calib (12.5%)
        # This gives us 60% train, 10% val, 10% calib, 20% test of original data
        x_train_temp, x_val_calib, y_train_temp, y_val_calib = train_test_split(
            x_train_val, y_train_val, test_size=0.25, random_state=seed)
        x_val, x_calib, y_val, y_calib = train_test_split(
            x_val_calib, y_val_calib, test_size=0.5, random_state=seed + 1)
        x_train, y_train = x_train_temp, y_train_temp
        
        # Create split_dict for conformalized version with separate calibration set
        split_dict = {
            "x_train": x_train.to_numpy(),
            "y_train": y_train,
            "x_val": x_val.to_numpy(),
            "y_val": y_val,
            "x_calib": x_calib.to_numpy(),
            "y_calib": y_calib,
            "x_test": x_test.to_numpy(),
            "y_test": y_test
        }
        worker_logger.log(f"[Seed {seed_index}] Conformalized mode: Train={len(y_train)}, Val={len(y_val)}, Calib={len(y_calib)}, Test={len(y_test)}")
    else:
        # Standard approach: reuse validation set as calibration set
        x_train, x_val, y_train, y_val = train_test_split(x_train_val, y_train_val, test_size=0.25, random_state=seed)
        
        # Create split_dict for standard version (val = calib)
        split_dict = {
            "x_train": x_train.to_numpy(),
            "y_train": y_train,
            "x_val": x_val.to_numpy(),
            "y_val": y_val,
            "x_calib": x_val.to_numpy(),  # Reuse validation set
            "y_calib": y_val,             # Reuse validation set
            "x_test": x_test.to_numpy(),
            "y_test": y_test
        }
        worker_logger.log(f"[Seed {seed_index}] Standard mode: Train={len(y_train)}, Val/Calib={len(y_val)}, Test={len(y_test)}")
    
    # Optimize GAM smoothing parameters using training data
    n_splines = 10
    
    # Define quantile estimators based on qxgb_only flag
    if qxgb_only:
        # Use only XGBoost for qxgb variant
        quantile_estimators = {
            "QXGB": XGBRegressor(random_state=777, objective='reg:quantileerror', n_estimators=100, tree_method='hist', quantile_alpha=0.5, min_child_weight = 10)
        }
        worker_logger.log(f"[Seed {seed_index}] QXGB-only mode: Using only XGBoost quantile estimator")
        # Set default values for GAM parameters when not using GAM
        optimal_lam = None
        gam_converged = False
    else:
        # Define basic quantile estimators first (will always have these)
        quantile_estimators = {
            "QRF": RandomForestQuantileRegressor(random_state=777, n_estimators=100, default_quantiles=0.5, min_samples_leaf = 10),
            "QXGB": XGBRegressor(random_state=777, objective='reg:quantileerror', n_estimators=100, tree_method='hist', quantile_alpha=0.5, min_child_weight = 10)
        }
        
        # Set timeout to 5 minutes (300 seconds)
        optimal_lam, gam_converged = optimize_gam_smoothing(split_dict["x_train"], split_dict["y_train"], n_splines=n_splines, timeout=300)
        
        if gam_converged:
            worker_logger.log(f"[Seed {seed_index}] ExpectileGAM converged with optimal smoothing parameter (lam): {optimal_lam:.4f}")
            # Add ExpectileGAM to the models if it converged successfully
            quantile_estimators["ExpectileGAM"] = ExpectileGAM(expectile=0.5, n_splines=n_splines, lam=optimal_lam, spline_order=3, basis='ps')
        else:
            worker_logger.log(f"[Seed {seed_index}] ExpectileGAM did not converge within timeout. Will use only {', '.join(quantile_estimators.keys())} models.")

    # Print the models we're using for this run
    worker_logger.log(f"[Seed {seed_index}] Using models: {', '.join(quantile_estimators.keys())}")
    
    # Create a dictionary to store results for each alpha
    all_results = {}
    
    # Process each alpha value
    for alpha in alphas:
        worker_logger.log(f"[Seed {seed_index}] Training and calibrating for alpha={alpha} (confidence level: {(1-alpha)*100}%)")
        
        # Create the appropriate PCS object
        if use_oob:
            worker_logger.log(f"[Seed {seed_index}] Using OOB method with top_k={top_k}")
            pcs_model = PCS_OOB(
                models=quantile_estimators,
                num_bootstraps=n_boot,
                alpha=alpha,
                seed=seed,
                top_k=top_k,
                calibration_method=calibration_method
            )
            # Fit the model on train+val combined data
            pcs_model.fit(
                X = np.concatenate([split_dict["x_train"], split_dict["x_val"]]), 
                y = np.concatenate([split_dict["y_train"], split_dict["y_val"]])
            )
        else:
            pcs_model = PCS_UQ(
                models=quantile_estimators,
                num_bootstraps=n_boot,
                alpha=alpha,
                seed=seed,
                top_k=top_k,
                calibration_method=calibration_method
            )
            pcs_model.fit(X = split_dict["x_train"], y = split_dict["y_train"], X_calib = split_dict["x_val"], y_calib = split_dict["y_val"])
        
        worker_logger.log(f"[Seed {seed_index}] Calibration complete. Gamma value: {pcs_model.gamma}")
        
        # Print model performance rankings to the worker logger
        print_model_performance(pcs_model, worker_logger)

        if use_oob:
            worker_logger.log(f"[Seed {seed_index}] Using OOB method with custom get_intervals")
            
            # Use get_intervals on train+val combined data
            train_intervals_raw = get_intervals_manually(
                pcs_model, 
                split_dict["x_train"],
                alpha=alpha
            )
            
            val_intervals_raw = get_intervals_manually(
                pcs_model, 
                split_dict["x_val"],
                alpha=alpha
            )
            
            calib_intervals_raw = get_intervals_manually(
                pcs_model, 
                split_dict["x_calib"],
                alpha=alpha
            )
            
            # Use the same method for test data directly
            test_intervals_raw = get_intervals_manually(pcs_model, split_dict['x_test'], alpha=alpha)
            
            # Get calibrated intervals
            train_intervals = pcs_model.predict(split_dict['x_train'])
            val_intervals = pcs_model.predict(split_dict['x_val'])
            calib_intervals = pcs_model.predict(split_dict['x_calib'])
            test_intervals = pcs_model.predict(split_dict['x_test'])
            
            worker_logger.log(f"[Seed {seed_index}] OOB intervals and medians calculated using custom method")
        else:
            # For standard PCS_UQ, use the built-in methods
            train_intervals_raw = pcs_model.get_intervals(split_dict['x_train'])
            val_intervals_raw = pcs_model.get_intervals(split_dict['x_val'])
            calib_intervals_raw = pcs_model.get_intervals(split_dict['x_calib'])
            test_intervals_raw = pcs_model.get_intervals(split_dict['x_test'])
            
            # Get calibrated intervals
            train_intervals = pcs_model.predict(split_dict['x_train'])
            val_intervals = pcs_model.predict(split_dict['x_val'])
            calib_intervals = pcs_model.predict(split_dict['x_calib'])
            test_intervals = pcs_model.predict(split_dict['x_test'])
            
            worker_logger.log(f"[Seed {seed_index}] Standard intervals and medians calculated")
        
        # Store results
        pcs_ensemble = {
            'x_train': split_dict['x_train'], 
            'y_train': split_dict['y_train'],
            'x_val': split_dict['x_val'], 
            'y_val': split_dict['y_val'],
            'x_calib': split_dict['x_calib'],
            'y_calib': split_dict['y_calib'],
            'x_test': split_dict['x_test'], 
            'y_test': split_dict['y_test'],
            # Raw intervals for all datasets
            'train_intervals_raw': train_intervals_raw,
            'val_intervals_raw': val_intervals_raw,
            'calib_intervals_raw': calib_intervals_raw,
            'test_intervals_raw': test_intervals_raw,
            # Calibrated intervals for all datasets
            'train_intervals': train_intervals,
            'val_intervals': val_intervals,
            'calib_intervals': calib_intervals,
            'test_intervals': test_intervals,
            # Median predictions for all datasets
            # 'train_median_preds': train_median_preds,
            # 'val_median_preds': val_median_preds,
            # 'test_median_preds': test_median_preds,
            'alpha': alpha,
            'confidence_level': f"{(1-alpha)*100:.0f}",
            'gamma': pcs_model.gamma,
            'top_k': top_k,
            'calibration_method': calibration_method,
            'use_oob': use_oob,
            'conformalized': conformalized,
            'top_model_names': list(pcs_model.top_k_models.keys()),
            'model_performances': {model: score for model, score in sorted(pcs_model.pred_scores.items(), key=lambda x: x[1], reverse=True)},
            'gam_parameters': {'lam': optimal_lam, 'n_splines': n_splines},
            'gam_converged': gam_converged
        }
        
        all_results[f"alpha_{alpha}"] = pcs_ensemble
    
    # Return both results and logs
    return {
        'results': all_results,
        'logs': worker_logger.get_logs(),
        'seed_index': seed_index
    }

def run_single_experiment_pcs_and_save(dataset_name, seed_index, n_boot=50, base_data_path=None, save_path_template=None, alphas=[0.1, 0.05, 0.01], top_k=1, use_oob=False, calibration_method='multiplicative', conformalized=False, qxgb_only=False):
    """Modified function that runs the experiment and saves result directly to disk"""
    full_result = run_single_experiment_pcs(dataset_name, seed_index, n_boot, base_data_path, alphas, top_k, use_oob, calibration_method, conformalized, qxgb_only)
    results = full_result['results']
    
    # Save results for each alpha to individual files
    if save_path_template:
        for alpha_key, result in results.items():
            alpha = result['alpha']
            confidence_level = result['confidence_level']
            save_path = save_path_template.format(f"{dataset_name}_run_{seed_index}_{confidence_level}")
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            
            # Convert data to a serializable format before saving
            serializable_result = convert_to_serializable(result)
            with open(save_path, "wb") as f:
                pickle.dump(serializable_result, f)
        
    return full_result

if __name__ == "__main__":
    # Default list of datasets
    default_datasets = [
        'data_computer',
        'data_ailerons',
        'data_airfoil',
        'data_ca_housing',
        'data_concrete',
        'data_elevator',
        'data_energy_efficiency',
        'data_insurance',
        'data_kin8nm',
        'data_miami_housing',
        'data_naval_propulsion',
        'data_parkinsons',
        'data_powerplant',
        'data_sulfur',
        'data_superconductor',
        'data_qsar',
        'data_allstate', # also takes long computationally, but we have maintained it
        # 'data_diamond', #disabled for computational reasons
    ]

    parser = argparse.ArgumentParser(description='Run PCS ensemble experiments')
    parser.add_argument('--datasets', type=str, nargs='?', default=','.join(default_datasets), 
                        help='Comma-separated list of dataset names (default: all datasets)')
    parser.add_argument('--seed_index', type=int, default=0, help='Seed index')
    parser.add_argument('--n_boot', type=int, default=100, help='Number of bootstrap samples')
    parser.add_argument('--n_jobs', type=int, default=5, help='Number of jobs for parallelizing the training')
    parser.add_argument('--project_root', type=str, default=os.path.join("..", ".."), help='Base path for the data')
    parser.add_argument('--alphas', type=str, default='0.05', 
                       help='Comma-separated list of alpha values for calibration (default: 0.1,0.05,0.01)')
    parser.add_argument('--top_k', type=int, default=1, help='Number of top-performing models to use (default: 1)')
    parser.add_argument('--use_oob', action='store_true', help='Use PCS_OOB instead of PCS_UQ')
    parser.add_argument('--calibration_method', type=str, choices=['multiplicative', 'additive'], 
                       default='multiplicative', help='Method for calibration (default: multiplicative)')
    parser.add_argument('--conformalized', action='store_true', 
                       help='Use separate calibration set (conformalized) instead of reusing validation set (standard)')
    parser.add_argument('--qxgb_only', action='store_true', 
                       help='Use only XGBoost quantile estimator (qxgb variant) instead of full PCS ensemble')
    parser.add_argument('--n_runs', type=int, default=10, help='Number of runs for each dataset (default: 5)')
    parser.add_argument('--log_file', type=str, default=None, 
                       help='Path to log file (default: logs/pcs_training_<timestamp>.log)')
    args = parser.parse_args()

    # Set up logging
    logger = setup_logging(args.log_file)
    logging.info("=" * 80)
    mode_type = "QXGB-only" if args.qxgb_only else "PCS ensemble"
    logging.info(f"Starting {mode_type} training with the following parameters:")
    logging.info(f"- Datasets: {args.datasets}")
    logging.info(f"- Seed index: {args.seed_index}")
    logging.info(f"- Bootstrap samples: {args.n_boot}")
    logging.info(f"- Jobs: {args.n_jobs}")
    logging.info(f"- Project root: {args.project_root}")
    logging.info(f"- Use OOB: {args.use_oob}")
    logging.info(f"- Calibration method: {args.calibration_method}")
    logging.info(f"- Conformalized: {args.conformalized}")
    logging.info(f"- QXGB only: {args.qxgb_only}")
    logging.info(f"- Top k: {args.top_k}")
    logging.info("=" * 80)

    # Parse the datasets from command-line input (or use the default list)
    datasets = args.datasets.split(',')
    
    # Parse alpha values
    alphas = [float(alpha) for alpha in args.alphas.split(',')]
    logging.info(f"Using alpha values: {alphas} (confidence levels: {[(1-alpha)*100 for alpha in alphas]}%)")

    data_base_path = os.path.join(args.project_root, "data")
    approach_suffix = "conformalized" if args.conformalized else "standard"
    # Naming convention:
    #  - QXGB-only:    models/pcs_top{top_k}_qxgb_{n_runs}_{approach}
    #  - PCS ensemble: models/pcs_top{top_k}_qpcs_{n_runs}_{approach}
    model_prefix = "qxgb" if args.qxgb_only else "qpcs"
    model_save_path = os.path.join(
        args.project_root,
        "models",
        f"pcs_top{args.top_k}_{model_prefix}_{args.n_runs}_{approach_suffix}",
        "{}_pcs_results_{}.pkl",
    )
    # Create results directory if it doesn't exist
    os.makedirs(os.path.dirname(model_save_path.format("", "")), exist_ok=True)

    logging.info("Starting to process datasets...")
    start_time = time.time()
    #---------------------------------------------------
    # Option 1: Process everything at once
    # Process each dataset
    for dataset_name in datasets:
        dataset_name = dataset_name.strip()
        logging.info(f"Processing dataset: {dataset_name}")

        # Special handling for Allstate due to convergence issues
        n_jobs_to_use = 1 if dataset_name == 'data_allstate' else args.n_jobs
        
        all_results = Parallel(n_jobs=n_jobs_to_use)(
            delayed(run_single_experiment_pcs)(
                dataset_name, i, 
                n_boot=args.n_boot, 
                base_data_path=data_base_path, 
                alphas=alphas,
                top_k=args.top_k,
                use_oob=args.use_oob,
                calibration_method=args.calibration_method,
                conformalized=args.conformalized,
                qxgb_only=args.qxgb_only
            )
            for i in range(args.n_runs)
        )
        
        # Print worker logs from each seed
        for result in all_results:
            seed_idx = result['seed_index']
            logging.info(f"\n----- Logs from worker process (seed {seed_idx}) -----")
            for log_line in result['logs']:
                logging.info(log_line)
        
        # For each alpha, create a separate results file
        for alpha in alphas:
            confidence_level = f"{(1-alpha)*100:.0f}"
            results_for_alpha = {f'run_{i}': all_results[i]['results'][f"alpha_{alpha}"] for i in range(args.n_runs)}
            
            # Convert to serializable before saving
            serializable_results = convert_to_serializable(results_for_alpha)
            
            # Save results
            with open(model_save_path.format(dataset_name, confidence_level), "wb") as f:
                pickle.dump(serializable_results, f)
            
            logging.info(f"Results saved for dataset {dataset_name} at confidence level {confidence_level}")

        # end time
        end_time = time.time()
        logging.info(f"Total time taken for {dataset_name}: {end_time - start_time} seconds")

    logging.info("All processing complete!")


# To run the experiments, use the following command (for our paper):
# python .\train_pcs_quantile.py --top_k 1
## Alternatively, for double-logging for each top_k, use the following command:
## python .\train_pcs_quantile.py --top_k 1 *> logs/pcs_training_top1.log ; python .\train_pcs_quantile.py --top_k 2 *> logs/pcs_training_top2.log